import uuid
from pydantic import BaseModel, ConfigDict
from typing import Dict
from logging import getLogger
from airbot_data_collection.basis import Bcolors
from airbot_data_collection.common.samplers.mcap_sampler import (
    McapDataSampler,
    McapDataSamplerConfig,
)


try:
    from dataloop import DataLoopClient

    DATALOOP_AVAILABLE = True
except ImportError:
    DATALOOP_AVAILABLE = False
    getLogger(__name__).warning(
        "It is detected that the `UPLOAD` package is not installed, and the cloud upload function will not be available. If you need to use the upload function, please contact us to install it.",
    )


class UploadConfig(BaseModel, frozen=True):
    model_config = ConfigDict(extra="forbid")

    enable: bool = False
    endpoint: str = ""
    username: str = ""
    password: str = ""


class AIRBOTMcapDataSamplerConfig(McapDataSamplerConfig):
    upload: UploadConfig = UploadConfig()


class AIRBOTMcapDataSampler(McapDataSampler):
    config: AIRBOTMcapDataSamplerConfig
    _info: Dict[str, Dict[str, str]]

    def on_configure(self):
        """Configure the airbot mcap data sampler."""
        self._init_upload()
        return super().on_configure()

    def save(self, path: str, data: dict) -> str:
        """Save the data to a MCAP file."""
        path = super().save(path, data)
        # Upload to cloud after saving
        if self.config.upload.enable:
            self._upload_to_cloud(path)
        return path

    def _init_upload(self):
        """Enable upload to cloud storage."""
        if self.config.upload.enable:
            assert DATALOOP_AVAILABLE, "DataLoopClient is not available"
            self.dataloop_client = DataLoopClient(
                endpoint=self.config.upload.endpoint,
                username=self.config.upload.username,
                password=self.config.upload.password,
            )
            self.get_logger().info(
                Bcolors.cyan(f"Will upload to task id: {self.config.task_info.task_id}")
            )

    def _upload_to_cloud(self, file_path: str) -> bool:
        """Upload the saved file to cloud storage."""
        # Convert task_id to int if it's a string
        project_id = self.config.task_info.task_id
        if isinstance(project_id, str):
            project_id = int(project_id)
        message = self.dataloop_client.samples.upload_sample(
            project_id=project_id,
            sample_id=str(uuid.uuid4()),
            sample_type="Sequential",
            file_path=file_path,
        )
        self.get_logger().info(Bcolors.green(f"Uploaded to cloud: {message}"))
        return True



from typing import Union, Optional
from airbot_data_collection.common.systems.basis import Sensor, System, SystemMode
from airbot_data_collection.common.demonstrators.basis import Demonstrator
from airbot_data_collection.demonstrate.configs import DemonstrateAction
from airbot_data_collection.common.systems.grouped import (
    GroupedComponentsSystemConfig,
    GroupedComponentsSystem,
    GroupsSendActionConfig,
)


Component = Union[System, Sensor]


class GroupedDemonstratorConfig(GroupedComponentsSystemConfig):
    """Configuration for GroupedDemonstrator"""


class GroupedDemonstrator(Demonstrator):
    config: GroupedDemonstratorConfig
    interface: GroupedComponentsSystem

    def on_configure(self):
        return self.interface.configure()

    def send_action(self, action: GroupsSendActionConfig) -> bool:
        """Control the leaders after some demonstrate action"""
        return self.interface.send_action(action)

    def capture_observation(self, timeout: Optional[float] = None):
        return self.interface.capture_observation(timeout)

    def get_info(self):
        return self.interface.get_info()

    def react(self, action):
        if action is DemonstrateAction.sample:
            return self.switch_mode(SystemMode.PASSIVE)
        elif action is DemonstrateAction.deactivate:
            return self.interface.handler.exit()
        elif action is DemonstrateAction.configure:
            return self.configure()
        else:
            func = getattr(self, f"_{action.value}", None)
            if func is not None:
                return func()
        return True

    def _activate(self) -> bool:
        # try to start auto control
        if self.interface.start_auto_control():
            # set leaders to resetting mode
            return self.switch_mode(SystemMode.RESETTING)

    def _finish(self) -> bool:
        return self.shutdown()

    def on_switch_mode(self, mode):
        return self.interface.switch_mode(mode)

    def shutdown(self) -> bool:
        return self.interface.shutdown()

    @property
    def handler(self):
        return self.interface.handler



from airbot_data_collection.common.devices.cameras.v4l2 import (
    V4L2Camera,
    V4L2CameraConfig,
)
from airbot_data_collection.common.visualizers.opencv import (
    OpenCVVisualizer,
    OpenCVVisualizerConfig,
)
from airbot_data_collection.common.devices.cameras.utils import (
    find_video_capture_devices,
)
from airbot_data_collection.utils import (
    init_logging,
    execute_shell_script,
    get_can_interfaces,
    zip,
    BaseModelWithFieldAliases,
)
from airbot_data_collection.common.utils.system_info import SystemInfo
from airbot_data_collection.basis import PACKAGE_NAME, Bcolors
from collections import defaultdict
from pprint import pformat
from pydantic import Field
from pydantic_settings import CliApp
from typing import List, Dict
from importlib.metadata import version
from pathlib import Path
from ruamel.yaml import YAML
import logging
import cv2
import subprocess
import time


yaml = YAML()
init_logging()
logger = logging.getLogger(f"{PACKAGE_NAME}-setup")
logger.info(f"Version: {version(PACKAGE_NAME)}")

try:
    from airbot_data_collection.common.devices.cameras.intelrealsense import (
        IntelRealSenseCamera,
        IntelRealSenseCameraConfig,
        find_camera_device_ids,
    )

    USE_REALSENSE = True
except ImportError as e:
    logger.warning(e)
    USE_REALSENSE = False


def list_to_nested_tuples(lst):
    return [(lst[i], lst[i + 1]) for i in range(0, len(lst), 2)]


def check_can_interfaces(expected_interfaces: list[str]) -> bool:
    result = subprocess.run(
        ["ip", "l"],
        capture_output=True,
        text=True,
        check=True,
    )
    output = result.stdout

    found_interfaces = [line for line in output.splitlines() if "can_" in line]
    found_names = [line.split(":")[1].strip() for line in found_interfaces]

    missing = [name for name in expected_interfaces if name not in found_names]
    if not missing:
        return True
    else:
        return False


cur_dir = Path(__file__).parent.resolve()


class SetupConfig(BaseModelWithFieldAliases):
    """Configuration for the setup script of airbot data collection."""

    ignore_cameras: List[str] = Field(
        [],
        validation_alias="ic",
        description="Ignore cameras by their bus_info or serial_number",
    )
    can_interfaces: List[str] = Field(
        [],
        validation_alias="can",
        description="List of CAN interfaces to use. If not provided, all available CAN interfaces will be used.",
    )
    ignore_cans: List[str] = Field(
        [],
        validation_alias="ii",
        description="Ignore CAN interfaces by their names.",
    )
    ref_cfg_dir: Path = Field(
        "airbot_ie/configs/demonstrators/",
        validation_alias="rcd",
        description="Directory containing the base configuration files.",
    )
    ref_cfg_name: Path = Field(
        "basis",
        validation_alias="rcn",
        description="Name of the base configuration file.",
    )


args = CliApp.run(SetupConfig)

logger.info("Getting system information...")
hw_uuid = SystemInfo.get_product(True).get("uuid", "unknown")
logger.info(f"Hardware uuid: {hw_uuid}")

"""Process Configs"""

ref_cfg_path = args.ref_cfg_dir / args.ref_cfg_name.with_suffix(".yaml")
if not ref_cfg_path.exists():
    raise FileNotFoundError(f"Base config file not found: {ref_cfg_path.absolute()}")

station_config_path = cur_dir / "station_config.yaml"
station_config = yaml.load(open(station_config_path))
NAME_CHOICES = station_config["choices"]
BUS_NAME_MAPPINGS = station_config["bus_name_mapping"]
# TODO: support for X5
CAN_NAME_MAPPINGS = {
    2: ["can_lead", "can_follow"],
    4: ["can_left_lead", "can_left", "can_right_lead", "can_right"],
}

"""Process CAN Interfaces"""

can_itfs = args.can_interfaces or sorted(
    set(get_can_interfaces()) - set(args.ignore_cans)
)
can_num = len(can_itfs)
assert can_num in BUS_NAME_MAPPINGS, f"Not correct can number: {can_itfs}"
can_buses = list_to_nested_tuples(can_itfs)
can_group_num = len(can_buses)
logger.info(f"CAN interfaces: {can_buses}")

if hw_uuid not in BUS_NAME_MAPPINGS[can_num]:
    BUS_NAME_MAPPINGS[can_num][hw_uuid] = {}
bus_name_mapping: dict = BUS_NAME_MAPPINGS[can_num][hw_uuid]
target_cans = CAN_NAME_MAPPINGS[can_num]
name_choices = NAME_CHOICES[can_num]

if set(target_cans) != set(can_itfs):
    logger.info(
        Bcolors.cyan(
            f"Binding CAN group {can_itfs} to target interfaces {target_cans}..."
        )
    )
    execute_shell_script(
        f"{cur_dir}/bind_can_udev.sh",
        args=[
            "--target",
            *target_cans,
        ],
        with_sudo=True,
    )
    # TODO: detect whether the CAN interfaces are bound correctly
    logger.info(
        Bcolors.cyan(
            "Please reconnect the robotic arms and press `Enter` to continue..."
        )
    )
    input()
    logger.info("Waiting for the system to stabilize after reconnection...")
    time.sleep(4)
    if check_can_interfaces(target_cans):
        logger.info(Bcolors.green("Successfully bound."))
    else:
        logger.error("Failed to bind. Please check the connections.")
        exit(1)
else:
    logger.info(f"CAN {can_itfs} already bound correctly.")

"""Process Cameras"""

all_cam_devices = find_video_capture_devices(True)
realsense_buses = []
logger.info(Bcolors.cyan(f"Found v4l2 devices: \n{pformat(all_cam_devices)}"))
for device_key in list(all_cam_devices.keys()):
    bus_id = device_key[1]
    if bus_id in args.ignore_cameras:
        all_cam_devices.pop(device_key)
    elif "RealSense" in device_key[0]:
        # remove realsense cameras from the v4l2 devices
        # since we will use their serial numbers
        all_cam_devices.pop(device_key)
        realsense_buses.append(bus_id)
used_camera_indices = [cam_ids[0] for cam_ids in all_cam_devices.values()]
# add realsene camera serial numbers
realsense_serials = set()
if USE_REALSENSE:
    realsense_cams = find_camera_device_ids(True)
    for bus, serial in realsense_cams.items():
        if bus not in args.ignore_cameras and serial not in realsense_serials:
            realsense_serials.add(serial)
    used_camera_indices.extend(realsense_serials)
else:
    logger.warning(
        f"Ignoring Intel RealSense cameras since the package is not installed: {realsense_buses}."
    )
    args.ignore_cameras.extend(realsense_buses)
assert used_camera_indices, "No used cameras. Please check the args and connections."
logger.info(Bcolors.blue(f"All ignored cameras: {args.ignore_cameras}"))
logger.info(Bcolors.blue(f"Used camera indices: {used_camera_indices}"))

cameras: list[V4L2Camera] = []
camera_vis_keys: list[str] = []
camera_bus_serials: list[str] = []
visualizers: list[OpenCVVisualizer] = []
camera_types: dict[str, str] = {}
camera_indices = []
camera_filenames = []
camera_params = defaultdict(dict)

cfged_indices = []
cfged_bus_serials = []
cfged_names = []
cfged_camera_types = []
no_cfg_buses_indexes: list[int] = []
for i, index in enumerate(list(used_camera_indices)):
    is_realsense = index in realsense_serials
    camera_config = {
        "width": 640,
        "height": 480,
    }
    if is_realsense:
        camera_config["fps"] = 30
        config = IntelRealSenseCameraConfig(
            camera_index=index, enable_depth=False, **camera_config
        )
        camera = IntelRealSenseCamera(config)
        camera_type = "realsense"
    else:
        config = V4L2CameraConfig(
            camera_index=index, pixel_format="MJPEG", decode=False, **camera_config
        )
        camera = V4L2Camera(config)
        camera_type = "v4l2"
    visualizer = OpenCVVisualizer(OpenCVVisualizerConfig(ignore_info=True, wait_key=-1))
    if camera.configure():
        if visualizer.configure():
            if is_realsense:
                bus = index
                file_name = camera_type
                camera_params[bus] = {
                    "fps": 30,
                }
                target = "airbot_data_collection.common.devices.cameras.intelrealsense.IntelRealSenseCamera"
            else:
                bus = camera.device.info.bus_info
                file_name = camera.device.filename
                target = "airbot_data_collection.common.devices.cameras.v4l2.V4L2Camera"
            camera_config["_target_"] = target
            camera_params[bus].update(camera_config)
            logger.info(f"Camera {index} bus/serial info: {bus}")
            prefix = bus_name_mapping.get(bus, "None")
            if prefix == "None":
                logger.info(
                    Bcolors.blue(
                        f"Camera {index} bus/serial info {bus} not found in bus/serial name mapping."
                    )
                )
                no_cfg_buses_indexes.append(i)
            else:
                cfged_bus_serials.append(bus)
                cfged_names.append(prefix)
                cfged_indices.append(index)
                cfged_camera_types.append(camera_type)
            vis_key = f"{prefix} : {file_name} : {bus}"
            cameras.append(camera)
            camera_vis_keys.append(vis_key)
            camera_bus_serials.append(bus)
            visualizers.append(visualizer)
            camera_indices.append(index)
            camera_types[bus] = camera_type
            camera_filenames.append(file_name)
        else:
            logger.error(f"Failed to configure visualizer for camera index {index}.")

if camera_indices:
    logger.info(f"Opened cameras: {camera_indices}")
else:
    logger.error("No camera opened. Please check the camera indices.")
    exit(1)

logger.info(
    Bcolors.blue(
        "\n"
        + pformat(
            {
                "q or ESC": "Quit the setup script without saving configs.",
                "c": "Configure cameras with names.",
                "s": "Save the current configuration and exit.",
            }
        )
        + "\nNote: Click any of the image windows and then press the key"
    )
)
while True:
    for camera, vis_key, visualizer in zip(cameras, camera_vis_keys, visualizers):
        obs = camera.capture_observation(2.0)
        if isinstance(obs, dict):
            obs = obs["color/image_raw"]
        visualizer.update({vis_key: obs}, None)
    key = cv2.waitKey(1) & 0xFF
    if key == ord("q") or key == 27:  # ESC or 'q' to quit
        logger.info("Exiting setup script.")
        break
    elif key == ord("c"):
        if not no_cfg_buses_indexes:
            logger.warning(
                "No need to configure since all cameras are mapped to their names"
            )
            logger.info(Bcolors.blue("Do you want to re-configue? (y/n)"))
            if cv2.waitKey(0) & 0xFF == ord("y"):
                no_cfg_buses_indexes = list(range(len(camera_bus_serials)))
                # clear the previous configuration
                cfged_names.clear()
                cfged_bus_serials.clear()
                cfged_indices.clear()
                cfged_camera_types.clear()
                camera_vis_keys.clear()
                for cam_fn, cam_bus_ser in zip(
                    camera_filenames,
                    camera_bus_serials,
                ):
                    camera_vis_keys.append(f"None : {cam_fn} : {cam_bus_ser}")
            else:
                continue
        unused_name = list(set(name_choices) - set(cfged_names))
        if len(unused_name) < len(no_cfg_buses_indexes):
            logger.error(
                f"Not enough names to configure cameras: {unused_name=} {no_cfg_buses_indexes=}."
                "Please remove the extra camera or update the `choices` field in the `station_config.yaml` config file with more name choices."
            )
            break
        logger.info(f"Configuring cameras {unused_name=} {cfged_names=}...")
        old_vis_keys = set()
        for bus_index in no_cfg_buses_indexes:
            bus = camera_bus_serials[bus_index]
            if len(unused_name) == 1:
                final_name = unused_name[0]
            else:
                hint_str = ""
                for i, name in enumerate(unused_name):
                    hint_str += f"{name}[{i}] | "
                logger.info(
                    Bcolors.cyan(
                        f"Name the camera on {bus} (press the digital number in []): {hint_str.removesuffix('| ')}"
                    )
                )
                index = cv2.waitKey(0) & 0xFF - ord("0")
                final_name = unused_name.pop(index)
            old_vis_key = camera_vis_keys[bus_index]
            # old_vis_keys.add(old_vis_key)
            camera_vis_keys[bus_index] = old_vis_key.replace("None", final_name)
            cfged_names.append(final_name)
            cfged_bus_serials.append(bus)
            cfged_indices.append(camera_indices[bus_index])
            cfged_camera_types.append(camera_types[bus])
            bus_name_mapping[bus] = final_name
            logger.info(Bcolors.green(f"Camera {bus} renamed to {final_name}"))
        with open(station_config_path, "w") as f:
            yaml.dump(station_config, f)
        logger.info(f"Updated station config: {station_config_path}")
        cv2.destroyAllWindows()
        no_cfg_buses_indexes.clear()
    elif key == ord("s"):
        if can_group_num == 1:
            groups = ["/"] * (len(can_itfs) + len(cfged_camera_types))
        elif can_group_num == 2:
            groups = ["left"] * 2 + ["right"] * 2 + ["/"] * len(cfged_camera_types)
        else:
            raise NotImplementedError(f"Not supported can group number {can_group_num}")

        components: Dict[str, list] = {
            "instances": [
                {
                    "_target_": "airbot_ie.robots.airbot_play.AIRBOTPlay",
                    "port": 50050 + i,
                }
                for i in range(len(can_itfs))
            ]
            + [
                {"camera_index": bus} | camera_params.get(bus, {})
                for bus in cfged_bus_serials
            ],
            "names": ["lead", "follow"] * can_group_num + cfged_names,
            "roles": ["l", "f"] * can_group_num + ["o"] * len(cfged_indices),
            "groups": groups,
        }
        ref_cfg_dir = ref_cfg_path.parent
        with open(ref_cfg_path) as f:
            config: dict = yaml.load(f)
            param_dict: dict = config["demonstrator"]["instance"]
            raw_comps = param_dict.get("components")
            if isinstance(raw_comps, dict):
                for key, value in raw_comps.items():
                    components[key].extend(value)
            logger.info(f"Components: {pformat(components)}")
            param_dict["components"] = components
            # config["defaults"].append(
            #     {"post_capture@demonstrator.instance": str(can_num)}
            # )
        # post_capture_path = f"{ref_cfg_dir}/post_capture/{can_num}.yaml"
        # with open(post_capture_path) as f:
        #     param_dict.update(yaml.load(f))
        with open(ref_cfg_dir / "setup.yaml", "w") as f:
            yaml.dump(config, f)
        logger.info(Bcolors.green("Setup completed successfully."))
        break
cv2.destroyAllWindows()

# disconnect all cameras,
# or the realsense camera may not be able to connect again1
logger.info("Shutting down cameras...")
for camera in cameras:
    camera.shutdown()



from typing import List, Union, Dict, Tuple, Any, Optional, Literal, Set
from collections.abc import Iterable
from pydantic import PositiveInt, Field
from time import time_ns, perf_counter
from collections import defaultdict
from functools import partial, cached_property
from airbot_data_collection.utils import linear_map, zip
from airbot_data_collection.common.systems.basis import (
    System,
    SystemConfig,
    InterfaceType,
    ReferenceMode,
    ActionConfigs,
    ObservationConfig,
    SystemMode,
)
from airbot_data_collection.common.utils.relative_control import RelativePoseControl
from airbot_data_collection.common.utils.coordinate import CoordinateTools
from airbot_data_collection.common.utils.tf import (
    apply_tf_to_pose,
    pose2matrix,
    to_matrix,
    array_pose_to_list_wrapper,
    is_identity_matrix,
    StaticTFBuffer,
)
from airbot_data_collection.common.configs.control import (
    JointControlBasis,
    JointPositionServo,
    JointPositionPlan,
    JointMIT,
    PoseControlBasis,
    PoseServo,
    PosePlan,
)
from mcap_data_loader.utils.basic import DictDataStamped, DataStamped
from functools import cache
import numpy as np


AVAILABLE_BACKEND = set()
try:
    from airbot_py.arm import AIRBOTArm, RobotMode, SpeedProfile

    AVAILABLE_BACKEND.add("grpc")
except ImportError:
    from airbot_ie.robots.airbot_play_thin import (
        AIRBOTArm,
        RobotMode,
        SpeedProfile,
    )

    AVAILABLE_BACKEND.add("thin")


ComponentType = Literal["arm", "eef"]


class AIRBOTPlayConfig(SystemConfig):
    url: str = "localhost"
    port: PositiveInt = 50050
    speed_profile: Optional[Union[SpeedProfile, str]] = SpeedProfile.FAST
    limit: Dict[str, Dict[Union[str, int], Tuple[float, float]]] = {}
    backend: str = "grpc"  # grpc or thin
    components: List[ComponentType] = Field(["arm", "eef"], min_length=1)
    action: ActionConfigs = [
        {
            SystemMode.RESETTING: JointPositionPlan(),
            SystemMode.SAMPLING: JointPositionServo(),
        }
    ]
    observation: List[ObservationConfig] = [
        ObservationConfig(
            interfaces=InterfaceType.joint_states() | {InterfaceType.POSE}
        )
    ]
    joint_names: List[List[str]] = [
        [f"joint{i}" for i in range(1, 7)],
        ["arm_eef_gripper_joint"],
    ]

    def model_post_init(self, context):
        if isinstance(self.speed_profile, str):
            self.speed_profile = SpeedProfile[self.speed_profile]
        assert self.backend in AVAILABLE_BACKEND, (
            f"Backend is not available: {self.backend}, "
            f"available backends: {AVAILABLE_BACKEND}"
        )

    @cached_property
    def pose_observation(self) -> bool:
        return InterfaceType.POSE in self.observation[0].interfaces

    @cached_property
    def relative_action(self) -> bool:
        return (
            self.action[0][SystemMode.SAMPLING].reference_mode != ReferenceMode.ABSOLUTE
        )

    @cached_property
    def relative_observation(self) -> bool:
        return self.observation[0].reference_mode != ReferenceMode.ABSOLUTE

    @cached_property
    def joint_fields(self) -> Set[str]:
        fields = set()
        prefix = "joint_"
        for interface in self.observation[0].interfaces:
            if interface.startswith(prefix):
                fields.add(interface.removeprefix(prefix))
        return fields


class AIRBOTPlay(System):
    interface: AIRBOTArm
    force_switch_mode: bool = False

    def __init__(self, config: AIRBOTPlayConfig):
        self.config = config
        self._joint_names = dict(zip(self.config.components, self.config.joint_names))

    def on_configure(self) -> bool:
        self._init_args()
        # mapping action config type to robot mode and function
        type2mode = {
            # NOTE: PLANNING_POS can be used for both joint and cartesian planning
            JointPositionPlan: RobotMode.PLANNING_POS,
            PosePlan: RobotMode.PLANNING_POS,
            JointPositionServo: RobotMode.SERVO_JOINT_POS,
            JointMIT: RobotMode.MIT_INTEGRATED,
            PoseServo: RobotMode.SERVO_CART_POSE,
        }
        type2func = {
            "arm": {
                JointPositionServo: self.interface.servo_joint_pos,
                JointPositionPlan: self.interface.move_to_joint_pos,
                JointMIT: self.interface.mit_joint_integrated_control,
                PoseServo: self.interface.servo_cart_pose,
                PosePlan: self.interface.move_to_cart_pose,
            },
            "eef": {
                JointPositionServo: self.interface.servo_eef_pos,
                JointPositionPlan: self.interface.move_eef_pos,
            },
        }
        # mapping action config type to action length
        type2length = {
            "arm": {
                JointPositionServo: 6,
                JointPositionPlan: 6,
                JointMIT: 0,
                PoseServo: 0,
                PosePlan: 0,
            },
            "eef": {
                JointPositionServo: 1,
                JointPositionPlan: 1,
                PoseServo: 0,
                PosePlan: 0,
            },
        }
        # mapping the system mode to robot mode
        self._mode_mapping = {SystemMode.PASSIVE: RobotMode.GRAVITY_COMP}
        mode_mapping = defaultdict(dict)
        self._mode2func = defaultdict(dict)
        self._mode2length = defaultdict(dict)
        config = self.config
        for component, mode_act_cfg in zip(config.components, config.action):
            for mode, act_cfg in mode_act_cfg.items():
                cfg_type = type(act_cfg)
                mode_mapping[mode][component] = type2mode[cfg_type]
                self._mode2func[mode][component] = type2func[component][cfg_type]
                self._mode2length[mode][component] = type2length[component][cfg_type]
        self._mode_mapping.update(mode_mapping)
        # print(f"mode_mapping: {self._mode_mapping}")
        # set action post process function TODO: configure this?
        self.action_post_process = self.action_data_to_list
        self.get_logger().info(f"Connecting to {config.url}:{config.port}")
        if self.interface.connect():
            # self.interface.set_speed_profile(self.config.speed_profile)
            self.interface.set_params(
                {
                    "servo_node.moveit_servo.scale.linear": 10.0,
                    "servo_node.moveit_servo.scale.rotational": 10.0,
                    "servo_node.moveit_servo.scale.joint": 1.0,
                    "sdk_server.max_velocity_scaling_factor": 1.0,
                    "sdk_server.max_acceleration_scaling_factor": 0.5,
                }
            )
            self._init_relative_control()
            # check if the robot components are available
            info = self.interface.get_product_info()
            self.get_logger().info(f"Robot info: {info}")
            info["arm_types"] = [info["product_type"]]
            for component in config.components:
                if info[f"{component}_types"][0] == "none":
                    self.get_logger().error(
                        f"Component {component} is not available. "
                        "Please check the configuration or the robot connection."
                    )
                    return False
                if component == "eef" and not self.interface.get_eef_pos():
                    self.get_logger().error(f"Can not get joint value of {component}")
                    return False
            return True
        return False

    @cache
    def _match_action_keys(
        self, mode: SystemMode, action_keys: Tuple[str]
    ) -> Dict[ComponentType, List[str]]:
        matched_keys = defaultdict(list)
        key_ends = {}
        for index, component in enumerate(self.config.components):
            # get the expected keys for each component
            act_cfg = self.config.action[index][mode]
            act_type = type(act_cfg)
            if issubclass(act_type, JointControlBasis):
                fields = ["position"]
                if act_type is JointMIT:
                    fields.extend(["velocity", "effort", "kp", "kd"])
                key_ends[component] = [
                    f"{component}/joint_state/{field}" for field in fields
                ]
            elif issubclass(act_type, PoseControlBasis):
                key_ends[component] = [
                    f"{component}/pose/position",
                    f"{component}/pose/orientation",
                ]
            else:
                raise ValueError(f"Unsupported action type: {act_type}")
        # print(f"Expected action key ends: {key_ends}")
        for component, key_ends in key_ends.items():
            for key_end in key_ends:
                # NOTE: eef pose should map to arm component
                replace = "eef" if "pose" in key_end else component
                for key in action_keys:
                    if key.replace(replace, component).endswith(key_end):
                        matched_keys[component].append(key)
                        break
        # print(f"Matched action keys: {matched_keys}")
        return dict(matched_keys)

    @staticmethod
    def action_data_to_list(action: DataStamped[np.ndarray]) -> List[float]:
        data = action["data"]
        if isinstance(data, list):
            return data
        return data.tolist()

    @staticmethod
    def action_to_list(action: np.ndarray) -> List[float]:
        return action.tolist()

    @staticmethod
    def action_forward(action: Any) -> Any:
        return action

    def send_action(
        self, action: Union[List[float], DictDataStamped[np.ndarray]]
    ) -> None:
        mode = self.current_mode
        component_func = self._mode2func[mode]
        if isinstance(action, dict):
            # tuple is hashable and can be cached
            act_keys = self._match_action_keys(mode, tuple(action.keys()))
            if not act_keys:
                self.get_logger().error(
                    f"No matching action keys from: {action.keys()}"
                )
                return
            for component, keys in act_keys.items():
                action_cfg = self.config.as_dict[component]["action"][mode]
                # flatten the action values
                if action_cfg.flatten:
                    target = []
                    for key in keys:
                        target.extend(self.action_post_process(action[key]))
                else:
                    target = [self.action_post_process(action[key]) for key in keys]
                    # TODO: is this always correct?
                    if len(target) == 1:
                        target = target[0]
                act_func = component_func[component]
                if action_cfg.unpack:
                    act_func(*target)
                else:
                    act_func(target)
        else:
            component_length = self._mode2length[self.current_mode]
            cnt = 0
            for component in self.config.components:
                length = component_length[component]
                act = action[cnt : cnt + length] if length > 0 else action[cnt]
                if act:  # error for numpy array
                    component_func[component](act)
                else:
                    break
                cnt += length or 1

    def on_switch_mode(self, mode: SystemMode) -> bool:
        # TODO: there
        robot_mode = self._mode_mapping[mode]
        if isinstance(robot_mode, RobotMode):
            robot_mode = {comp: robot_mode for comp in self.config.components}
        return self.interface.switch_mode(robot_mode["arm"])

    def _get_tf_key(self, component: str, frame: str) -> str:
        return f"{component}.{frame}"

    def _init_args(self):
        self._js_fields = self.config.joint_fields
        self._pose_fields = ("position", "orientation")
        self._post_capture = defaultdict(dict)
        limits: Dict[str, Dict[str, Dict[int, Tuple]]] = {
            "E2B": {"eef/joint_state/position": {0: (0, 0.0471)}},
            "G2": {
                "eef/joint_state/position": {0: (0, 0.0720)},
            },
            "play_pro": {
                "arm/joint_state/position": {0: (-2.74, 2.74)},
            },
            "play": {
                "arm/joint_state/position": {0: (-3.151, 2.080)},
            },
        }
        limits.update(
            {
                "PE2": limits["E2B"],
                "old_G2": limits["G2"],
                "play_lite": limits["play_pro"],
            }
        )
        self._default_limit = limits
        self._default_range = {
            "G2": {"eef/joint_state/position": {0: [0, 0.072]}},
            "play": {"arm/joint_state/position": {0: [-3.151, 2.080]}},
            "play_pro": {"arm/joint_state/position": {0: [-2.74, 2.74]}},
        }
        iden_rela_pose = ((0, 0, 0), (0, 0, 0, 1))
        x_pos = lambda x: (x, 0, 0)  # noqa: E731
        tf_dict = {
            "play": {
                "none": x_pos(0.0864995),
                "G2": x_pos(0.2466995),
                "old_G2": x_pos(0.2466995),
                "E2B": x_pos(0.1488995),
            },
        }
        tf_list = [("replay.PE2", "play.E2B", iden_rela_pose)]
        for arm_type, pos_rela in tf_dict.items():
            for eef_type, tf_part in pos_rela.items():
                tf_list.append(
                    (
                        self._get_tf_key(arm_type, eef_type),
                        self._get_tf_key(arm_type, "ref"),
                        tf_part,
                    )
                )
        self._tf_buffer = StaticTFBuffer(
            [(tgt, src, to_matrix(tf_part)) for tgt, src, tf_part in tf_list]
        )

    def _init_relative_control(self):
        pose = self.interface.get_end_pose()
        if self.config.relative_action:
            # TODO: support absolute mode instead of complex judgment
            self.rela_act_ctrl = RelativePoseControl(
                self.config.action[0].reference_mode
            )
            self.rela_act_ctrl.update(*pose)
        if self.config.relative_observation:
            self.rela_obs_ctrl = RelativePoseControl()
            self.rela_obs_ctrl.update(*pose)

    def _process_pose(
        self, pose: Union[List[float], List[list[float]]]
    ) -> List[list[float]]:
        # self.get_logger().info(f"Processing pose: {pose}")
        if not isinstance(pose[0], Iterable):
            pose = [pose[:3], pose[3:7]]

        if self.config.action[0].pose_reference_frame == "eef":
            cur_pose = self.interface.get_end_pose()
            pose = CoordinateTools.to_world_coordinate(pose, cur_pose)
        elif self.config.relative_action:
            if self.config.action[0].reference_mode.is_delta():
                self.rela_act_ctrl.update(*self.interface.get_end_pose())
            pose = self.rela_act_ctrl.to_absolute(*pose)

        # self.get_logger().info(f"Processed pose: {pose}")
        return [list(pose[0]), list(pose[1])]

    def capture_observation(
        self, timeout: Optional[float] = None
    ) -> dict[str, dict[str, Union[float, Dict[str, List[float]]]]]:
        """key: component_name/data_type"""
        obs = {}
        config = self.config
        if config.pose_observation:
            start = perf_counter()
            pose = self.interface.get_end_pose()
            # TODO: arm/pose?
            prefix = "eef/pose"
            # print(f"Raw pose: {pose}")
            pose = self._post_capture.get(prefix, lambda *args: args)(*pose)
            # print(f"Post processed pose: {pose}")
            if config.relative_observation:
                pose = self.rela_obs_ctrl.to_relative(*pose)
            for key, value in zip(self._pose_fields, pose):
                obs[f"{prefix}/{key}"] = {"t": time_ns(), "data": value}
            self._metrics["durations"]["capture/pose"] = perf_counter() - start
        start = perf_counter()
        for component in config.components:
            for field in self._js_fields:
                obs[f"{component}/joint_state/{field}"] = {
                    "t": time_ns(),
                    "data": self._get_joint_state(component, field),
                }
        self._metrics["durations"]["capture/joint_state"] = perf_counter() - start
        return obs

    def _get_joint_state(self, component: str, field: str) -> List[float]:
        if component == "eef" and field == "velocity":
            return [0.0] * len(self._joint_names[component])
        if field == "name":
            return self._joint_names[component]
        else:
            data = getattr(
                self.interface, f"get_{component.replace('arm', 'joint')}_{field[:3]}"
            )()
            for index, process in self._post_capture.get(
                f"{component}/joint_state/{field}", {}
            ).items():
                # self.get_logger().info(
                #     f"Processing {component}/joint_state/{field} at index {index}: {data[index]}"
                # )
                data[index] = process(data[index])
                # self.get_logger().info(f"Post value: {data[index]}")
            return data

    def shutdown(self) -> bool:
        return self.interface.disconnect()

    def get_info(self):
        return {
            key: list(value) if not isinstance(value, (str, bool)) else value
            for key, value in self.interface.get_product_info().items()
        } | {f"{comp}/joint_names": names for comp, names in self._joint_names.items()}

    def _get_default(
        self, arm_type: str, eef_type: str, default: dict
    ) -> Dict[str, dict]:
        return default.get(arm_type, {}) | default.get(eef_type, {})

    def set_post_capture(self, config, info):
        self_info = self.interface.get_product_info()
        arm_type = self_info["product_type"]
        eef_type = self_info["eef_types"][0]
        default_limits = self._get_default(arm_type, eef_type, self._default_limit)
        default_range = self._get_default(arm_type, eef_type, self._default_range)
        default_transf = {}
        if config is None or config.transform is None or config.transform:
            default_transf["eef/pose"] = self._tf_buffer.lookup_transform(
                self._get_tf_key(info["product_type"], info["eef_types"][0]),
                self._get_tf_key(arm_type, eef_type),
            )
        if config is None:
            range_mapping = default_range
            transform = default_transf
        else:
            range_mapping = config.range_mapping
            transform = {
                key: pose2matrix(*value) if value is not None else default_transf[key]
                for key, value in config.transform.items()
            }
        for key, value in range_mapping.items():
            limit = self.config.limit.get(key, {})
            default_limit = default_limits.get(key, {})
            default_limit.update(limit)
            try:
                for index, target_range in value.items():
                    self._post_capture[key][int(index)] = partial(
                        linear_map,
                        raw_range=default_limit[index],
                        target_range=target_range,
                    )
                    # self.get_logger().info(
                    #     f"Post capture config set: {target_range=}"
                    # )
            except Exception as e:
                raise RuntimeError(
                    f"Failed to set post capture for {key} with "
                    f"{default_limits=}, {arm_type=}, {eef_type=}"
                ) from e
        for key, value in transform.items():
            # print(f"tf_matrix={value}")
            self._post_capture[key] = (
                array_pose_to_list_wrapper(apply_tf_to_pose, tf_matrix=value)
                if not is_identity_matrix(value)
                else lambda *args: args
            )
        # self.get_logger().info(f"Post capture config set: {self._post_capture}")



from airbot_data_collection.basis import ConcurrentMode
from airbot_data_collection.common.utils.progress import (
    Waitable,
    ConcurrentProgressHandler,
)
from airbot_data_collection.utils import StrEnum
from multiprocessing import get_context
from enum import auto
from pydantic import NonNegativeInt, BaseModel
import time


class DemonstrateAction(StrEnum):
    configure = auto()
    activate = auto()
    capture = auto()
    sample = auto()
    update = auto()
    save = auto()
    remove = auto()
    abandon = auto()
    deactivate = auto()
    finish = auto()


class DemonstrateState(StrEnum):
    error = auto()
    unconfigured = auto()
    inactive = auto()
    active = auto()
    sampling = auto()
    finalized = auto()


SpawnEvent = get_context("spawn").Event


class SampleInfo(BaseModel):
    """Information of the current sampling round."""

    # NOTE: no frozen, no validation for performance consideration
    index: NonNegativeInt = 0
    """the current index (number) of the data in a single sample round"""
    round: NonNegativeInt = 0
    """the current round of the sampling"""


if __name__ == "__main__":
    from airbot_data_collection.utils import init_logging

    init_logging()

    handler = ConcurrentProgressHandler(ConcurrentMode.thread)

    def auto_control(waitable: Waitable):
        time.sleep(2)
        with waitable:
            while waitable.wait():
                print("Auto control is running...")
                time.sleep(1)
            print("Auto control has stopped.")

    handler.launch(target=auto_control, args=(handler.get_waitable(),), daemon=False)
    input("Press Enter to start...")
    handler.start()
    input("Press Enter to stop...")
    handler.stop()
    input("Press Enter to exit...")
    handler.exit()



from mcap_data_loader.datasets.mcap_dataset import (
    McapFlatBuffersSampleDataset,
    McapDatasetConfig,
)
from airbot_data_collection.common.systems.data_player import (
    DataPlayerConfig,
    IterablePlayer,
)
from typing import Dict, Union, Any, Optional
import numpy as np


class McapPlayerConfig(DataPlayerConfig):
    source: McapDatasetConfig


class McapPlayer(IterablePlayer):
    """A system that plays back data from an MCAP dataset."""

    config: McapPlayerConfig
    interface: McapFlatBuffersSampleDataset

    def capture_observation(
        self, timeout: Optional[float] = None
    ) -> Dict[str, Union[np.ndarray, Any]]:
        return super().capture_observation(timeout)



import numpy as np


class ImageCoder:
    @staticmethod
    def rgb2yuv(rgb: np.ndarray) -> np.ndarray:
        # The coefficients were taken from OpenCV https://github.com/opencv/opencv
        # I'm not sure if the values should be clipped, in my (limited) testing it looks alright
        #   but don't hesitate to add rgb.clip(0, 1, rgb) & yuv.clip(0, 1, yuv)
        #
        # Input for these functions is a numpy array with shape (height, width, 3)
        # Change '+= 0.5' to '+= 127.5' & '-= 0.5' to '-= 127.5' for values in range [0, 255]

        m = np.array(
            [
                [0.29900, -0.147108, 0.614777],
                [0.58700, -0.288804, -0.514799],
                [0.11400, 0.435912, -0.099978],
            ]
        )
        yuv = np.dot(rgb, m)
        yuv[:, :, 1:] += 127.5
        return yuv

    @staticmethod
    def yuv2rgb(yuv: np.ndarray) -> np.ndarray:
        # The coefficients were taken from OpenCV https://github.com/opencv/opencv
        # I'm not sure if the values should be clipped, in my (limited) testing it looks alright
        #   but don't hesitate to add rgb.clip(0, 1, rgb) & yuv.clip(0, 1, yuv)
        #
        # Input for these functions is a numpy array with shape (height, width, 3)
        # Change '+= 0.5' to '+= 127.5' & '-= 0.5' to '-= 127.5' for values in range [0, 255]

        m = np.array(
            [
                [1.000, 1.000, 1.000],
                [0.000, -0.394, 2.032],
                [1.140, -0.581, 0.000],
            ]
        )
        yuv[:, :, 1:] -= 127.5
        rgb = np.dot(yuv, m)
        return rgb

    @staticmethod
    def yuyv2bgr(data: bytes, width: int, height: int) -> np.ndarray:
        """Convert YUYV image bytes data to BGR array using numpy.
        Args:
            data: bytes of YUYV image
            height: image height
            width: image width
        Returns:
            numpy image array in bgr
        """
        yuyv = np.frombuffer(data, np.uint8).reshape((height, width, 2))
        y = yuyv[:, :, 0].astype(np.float32)
        u = np.zeros((height, width), dtype=np.float32)
        v = np.zeros((height, width), dtype=np.float32)
        u[:, 0::2] = yuyv[:, 0::2, 1]
        u[:, 1::2] = u[:, 0::2]
        v[:, 1::2] = yuyv[:, 1::2, 1]
        v[:, 0::2] = v[:, 1::2]
        r = y + 1.403 * (v - 128)
        g = y - 0.344 * (u - 128) - 0.714 * (v - 128)
        b = y + 1.770 * (u - 128)
        bgr = np.zeros((height, width, 3), dtype=np.uint8)
        bgr[:, :, 0] = np.clip(b, 0, 255).astype(np.uint8)  # B
        bgr[:, :, 1] = np.clip(g, 0, 255).astype(np.uint8)  # G
        bgr[:, :, 2] = np.clip(r, 0, 255).astype(np.uint8)  # R
        return bgr



import numpy as np
from typing import List, Any, Union
from pydantic import BaseModel, NonNegativeInt
from airbot_data_collection.common.utils.event_rpc import (
    EventRpcManager,
    ConcurrentMode,
    EventRpcServer,
)
from airbot_data_collection.common.wrappers.basis import WrapperBasis


class TemporalEnsemblingWithDroppingConfig(BaseModel):
    drop_num: NonNegativeInt = 0
    # empty list means average, float means exp decay
    weights: Union[List[float], float] = []
    max_steps: NonNegativeInt = 0


class TemporalEnsemblingWithDropping(WrapperBasis):
    """Temporal Ensembling for output time serial data with dropping method"""

    config: TemporalEnsemblingWithDroppingConfig

    def on_configure(self) -> bool:
        # TODO: use a dynamic method to adjust the
        # buffer size instead of allocating a very
        # large memory
        if self.config.max_steps == 0:
            self.config.max_steps = 2048
        if self.config.drop_num > 0:
            self.should_take_over_env = True
            mode = ConcurrentMode.thread
            self._rpc_manager = EventRpcManager(EventRpcManager.get_args(mode))
            self._call_thread = self._rpc_manager.get_concurrent_cls(mode)(
                target=self._async_call, args=(self._rpc_manager.server,), daemon=True
            )
            self._call_thread.start()
        return True

    def on_warm_up(self, output: Any):
        self._horizon = output.shape[1]
        if self.config.drop_num >= self._horizon:
            raise ValueError(
                f"The `drop_num`: {self.config.drop_num} must be less than the output horizon {self._horizon}"
            )
        self._ele_shape = output.shape[2:]
        weights = self.config.weights
        if isinstance(weights, float):
            weights = np.exp(-weights * np.arange(self._horizon))
        else:
            if not weights:
                weights = np.ones(self._horizon)
            elif len(weights) != self._horizon:
                raise ValueError(
                    f"The length of the weights {weights} must be equal to {self._horizon}"
                )
            weights = np.array(weights)
        if self.output_type_name == "Tensor":
            weights = (
                self.output_backend.from_numpy(weights)
                .to(device=self.output_device)
                .unsqueeze(dim=1)  # add a batch dim
            )
        self._weights = weights / weights.sum()
        # print(self._weights.shape)
        return output

    def on_reset(self):
        self._t = 0
        self._async_t = 0
        self._call_t = 0
        kwds = {}
        if self.output_type == "Tensor":
            kwds = {"device": self.output_device}
        self._all_time_outputs = self.output_backend.zeros(
            (self.config.max_steps, self.config.max_steps + self._horizon)
            + self._ele_shape,
            **kwds,
            dtype=self.output_dtype,
        )
        self.get_logger().info(
            f"Output buffer shape: {self._all_time_outputs.shape}, size: {self.calculate_size(self._all_time_outputs):.2f} MB"
        )

    def _async_call(self, rpc_server: EventRpcServer):
        self.get_logger().info("Call thread started")
        while rpc_server.wait():
            output = self.env.output()
            self.last_env_output = output
            self._sync_call(output.observation)
            rpc_server.respond()
        self.get_logger().info("Call thread exited")

    def _sync_call(self, *args, **kwds):
        t = self._async_t
        outputs = self.caller(*args, **kwds)
        # x axis uses t, y axis uses call_t
        self._all_time_outputs[[self._call_t], t : t + self._horizon] = outputs
        self._call_t += 1

    def call(self, *args, **kwds):
        t = self._t
        # need call
        if t == 0 or self.config.drop_num == 0 or ((t - 1) % self.config.drop_num == 0):
            self._async_t = t
            if self.taken_over_env:
                client = self._rpc_manager.client
                if not client.is_responded():
                    self.get_logger().info("Waiting for call thread to respond...")
                if client.wait(5.0):
                    if not client.request(0):
                        raise RuntimeError("Failed to request")
                    if t == 0:
                        # wait for the first call to be done
                        # to get the initial output
                        client.wait(5.0)
                else:
                    raise TimeoutError("Timeout waiting for call thread to be ready")
            else:
                self._sync_call(*args, **kwds)

        drop_num = self.config.drop_num
        hori = self._horizon
        eq_drop_num = 1 if drop_num == 0 else drop_num
        start = max(0, (t - (hori - eq_drop_num) - 1)) // eq_drop_num + (t > hori - 1)
        end = max(0, (t - 1) // drop_num if drop_num > 0 else t) + 1
        used_outputs = self._all_time_outputs[start:end, t]
        used_weights = self._weights[: end - start]
        # print(used_outputs)
        self._t += 1
        return (used_outputs * used_weights).sum(dim=0, keepdim=True)

    def on_shutdown(self):
        if self.taken_over_env:
            if self._rpc_manager.shutdown():
                self._call_thread.join(5.0)
                return not self._call_thread.is_alive()
        return True



if __name__ == "__main__":
    import argparse
    from airbot_data_collection.tests.test_mcap_player import McapSinglePosePlayer
    from airbot_ie.robots.airbot_play import (
        AIRBOTPlay,
        AIRBOTPlayConfig,
        ActionConfig,
        InterfaceType,
        SystemMode,
    )
    from airbot_data_collection.common.utils.transformations import (
        quaternion_from_euler,
    )
    import time
    import numpy as np

    parser = argparse.ArgumentParser(
        description="Play a single pose from an MCAP file."
    )
    parser.add_argument(
        "file_path",
        type=str,
        help="Path to the MCAP file containing the pose data.",
    )
    args = parser.parse_args()

    test = McapSinglePosePlayer(args.file_path)
    airbot_play = AIRBOTPlay(
        AIRBOTPlayConfig(
            port=50051, action=[ActionConfig(interfaces={InterfaceType.POSE})]
        )
    )
    assert airbot_play.configure()

    test.set_pose_bias(
        position=np.array([0.0, 0.0, 0.0]),
        orientation=quaternion_from_euler(0.0, 0.0, 0.0),
    )
    # test.set_eef_threshold(0.03, 0.0, 0.072)
    for i in range(1):
        test.seek(0)
        input(f"Press Enter to move to starting pose for iteration {i}...")
        assert airbot_play.switch_mode(SystemMode.RESETTING)
        airbot_play.send_action(test.update())
        input(f"Press Enter to start iteration {i}...")
        assert airbot_play.switch_mode(SystemMode.SAMPLING)
        period = 1 / 20.0
        while True:
            start = time.perf_counter()
            action = test.update()
            if not action:
                break
            print(action)
            airbot_play.send_action(action)
            time_elapsed = time.perf_counter() - start
            sleep_time = period - time_elapsed
            if sleep_time > 0:
                time.sleep(sleep_time)
        print(f"Finished iteration {i}.")
    assert airbot_play.shutdown()
    print("Finished iterating through the dataset.")



from airbot_ie.teleoprators.vr_quest import (
    VRQuestController,
    TeleopConfig,
    EventConfig,
)
from airbot_data_collection.common.devices.vr.pico import (
    VRPico,
    VRQuestConfig,
    VREvent,
)


class VRPicoController(VRQuestController):
    """
    Controller for the Pico VR headset, inheriting from VRQuestController.
    This class is designed to handle the specific events and configurations
    for the Pico VR headset, while reusing the functionality of the VRQuestController.
    """

    def _init_vr(self):
        self._vr = VRPico(
            VRQuestConfig(
                zero_info={comp: self._event_config.zero_info for comp in self._pos}
            )
        )


if __name__ == "__main__":
    from airbot_ie.teleoprators.vr_quest import main

    main(
        VRPicoController(
            TeleopConfig(
                event_config=EventConfig(
                    zero_info=VREvent.RIGHT_GRIP,
                    success=VREvent.LEFT_SECONDARY_BUTTON,
                    failure=VREvent.LEFT_PRIMARY_BUTTON,
                    rerecord_episode=VREvent.LEFT_GRIP,
                    intervention=VREvent.RIGHT_GRIP,
                    shutdown=VREvent.RIGHT_SECONDARY_BUTTON,
                    left_eef=VREvent.LEFT_STICK_V,
                    right_eef=VREvent.RIGHT_STICK_V,
                ),
            )
        )
    )



from airbot_data_collection.common.devices.cameras.utils import RGBDCameraConfig
from airbot_data_collection.common.systems.basis import Sensor, DictDataStamped
from airbot_data_collection.basis import force_set_attr
from typing import Optional
from time import time_ns
import numpy as np


class MockCameraConfig(RGBDCameraConfig):
    """Configuration for a mock camera device used for testing purposes."""

    random: bool = False

    @force_set_attr
    def model_post_init(self, context):
        if self.width is None:
            self.width = 640
        if self.height is None:
            self.height = 480


class MockCamera(Sensor):
    """A mock camera device used for testing purposes."""

    config: MockCameraConfig

    def on_configure(self):
        self._update_random_image()
        return True

    def capture_observation(
        self, timeout: Optional[float] = None
    ) -> DictDataStamped[np.ndarray]:
        if self.config.random:
            self._update_random_image()
        observation = {}
        if self.config.enable_color:
            observation["color/image_raw"] = {"t": time_ns(), "data": self.color_image}
        if self.config.enable_depth:
            if self.config.align_depth:
                key = "aligned_depth_to_color/image_raw"
            else:
                key = "depth/image_rect_raw"
            observation[key] = {"t": time_ns(), "data": self.depth_image}
        return observation

    def get_info(self):
        return self.config.model_dump(mode="json")

    def shutdown(self) -> bool:
        self.color_image = None
        self.depth_image = None
        return True

    def _update_random_image(self):
        if self.config.enable_color:
            self.color_image = np.random.randint(
                0, 256, (self.config.height, self.config.width, 3), dtype=np.uint8
            )
        if self.config.enable_depth:
            self.depth_image = np.random.randint(
                0,
                5000,
                (self.config.height, self.config.width),
                dtype=np.uint16,
            )



"""
Make one arm follow the movement of another arm.
"""

import logging

LOG_FORMAT = (
    "[%(asctime)s] [%(levelname)-8s] [%(name)s.%(funcName)s:%(lineno)d] - %(message)s"
)

# Configure the logging
logging.basicConfig(
    level=logging.DEBUG,  # Set the minimum logging level
    format=LOG_FORMAT,  # Use the custom format
    datefmt="%Y-%m-%d %H:%M:%S",  # Set the date format
    handlers=[logging.StreamHandler()],  # Output logs to stdout (console)
)


def follow():
    """
    Follow the lead robot arm
    """
    import argparse
    import time
    from airbot_py.arm import AIRBOTPlay, RobotMode, SpeedProfile
    from typing import List, Tuple

    logger = logging.getLogger(__name__)
    parser = argparse.ArgumentParser(
        description="Multi-pair lead/follow control example"
    )
    parser.add_argument(
        "--lead-url",
        type=str,
        nargs="+",
        help="URL(s) to connect to the lead server(s), space separated for multiple.",
    )
    parser.add_argument(
        "--follow-url",
        type=str,
        nargs="+",
        help="URL(s) to connect to the follow server(s), space separated for multiple.",
    )
    parser.add_argument(
        "-lp",
        "--lead-port",
        type=int,
        nargs="+",
        required=True,
        help="Server port(s) for arm(s) to lead, space separated for multiple.",
    )
    parser.add_argument(
        "-fp",
        "--follow-port",
        type=int,
        nargs="+",
        required=True,
        help="Server port(s) for arm(s) to follow, space separated for multiple.",
    )

    args = parser.parse_args()
    n = len(args.lead_port)
    if not args.lead_url:
        args.lead_url = ["localhost"] * n
    if not args.follow_url:
        args.follow_url = ["localhost"] * n
    if not (
        len(args.lead_url)
        == len(args.lead_port)
        == len(args.follow_url)
        == len(args.follow_port)
    ):
        raise ValueError("The number of lead/follow urls and ports must be the same.")
    for i in range(n):
        if (
            args.lead_port[i] == args.follow_port[i]
            and args.lead_url[i] == args.follow_url[i]
        ):
            raise ValueError(
                f"Pair {i}: Lead and follow port, lead and follow url cannot be the same at the same time."
            )

    robots: List[Tuple[AIRBOTPlay, AIRBOTPlay]] = []
    for i in range(n):
        lead = AIRBOTPlay(url=args.lead_url[i], port=args.lead_port[i])
        follow = AIRBOTPlay(url=args.follow_url[i], port=args.follow_port[i])
        print(
            f"Connecting to lead {args.lead_url[i]}:{args.lead_port[i]} and follow {args.follow_url[i]}:{args.follow_port[i]}"
        )
        assert lead.connect()
        assert follow.connect()
        robots.append((lead, follow))

    try:
        for lead, follow in robots:
            lead.switch_mode(RobotMode.PLANNING_POS)
            if (
                sum(
                    abs(a - b)
                    for a, b in zip(lead.get_joint_pos(), follow.get_joint_pos())
                )
                > 0.1
            ):
                follow.switch_mode(RobotMode.PLANNING_POS)
                follow.move_to_joint_pos(lead.get_joint_pos())
                time.sleep(1)
            logger.info(f"Lead robot arm {lead} is ready to follow.")
        for lead, follow in robots:
            lead.switch_mode(RobotMode.GRAVITY_COMP)
            follow.switch_mode(RobotMode.SERVO_JOINT_POS)
            # follow.set_speed_profile(SpeedProfile.FAST)
            follow.set_params(
                {
                    "servo_node.moveit_servo.scale.linear": 10.0,
                    "servo_node.moveit_servo.scale.rotational": 10.0,
                    "servo_node.moveit_servo.scale.joint": 1.0,
                    "sdk_server.max_velocity_scaling_factor": 1.0,
                    "sdk_server.max_acceleration_scaling_factor": 0.5,
                }
            )
        factor = 0.072 / 0.0471
        while True:
            for lead, follow in robots:
                follow.servo_joint_pos(lead.get_joint_pos())
                eef_pos = lead.get_eef_pos()
                if eef_pos:
                    eef_pos[0] *= factor
                else:
                    logger.info("Failed to get lead eef pos")
                follow.servo_eef_pos(eef_pos or [])
            time.sleep(0.01)
    finally:
        for lead, follow in robots:
            lead.switch_mode(RobotMode.PLANNING_POS)
            follow.switch_mode(RobotMode.PLANNING_POS)
            follow.set_speed_profile(SpeedProfile.DEFAULT)
        for lead, follow in robots:
            assert lead.disconnect()
            assert follow.disconnect()


if __name__ == "__main__":
    follow()



from mcap_data_loader.serialization.flb import (
    McapFlatBuffersWriter,
    FlatBuffersSchemas,
)
from airbot_data_collection.common.samplers.mcap_sampler import (
    McapDataSampler,
    McapDataSamplerConfig,
    McapTool,
    MediaType,
)
from airbot_data_collection.common.samplers.basis import TaskInfo
from mcap.writer import Writer
from pydantic import BaseModel
from pydantic_settings import CliApp
from typing import Dict
import os
import time
import json


class Config(BaseModel):
    """Configuration for the DISCOVERSE to MCAP conversion.
    Args:
        root (str): Root directory containing the task data.
        task_name (str): Name of the task to process.
        output_dir (str): Directory to save the output MCAP files. If not provided,
            it defaults to `<root>/mcap/<task_name>`.
    """

    root: str
    task_name: str
    output_dir: str = ""


config = CliApp.run(Config)

start = time.perf_counter()
directory = f"{config.root}/{config.task_name}"
output_dir = config.output_dir or f"{config.root}/mcap/{config.task_name}"

os.makedirs(output_dir, exist_ok=True)

# find all folders in the directory
folders = [f.path for f in os.scandir(directory) if f.is_dir()]
# print(folders)

config = McapDataSamplerConfig(task_info=TaskInfo(task_name=config.task_name))

for folder in folders:
    fd_base = os.path.basename(folder)
    if not fd_base.isdigit():
        print(f"Skipping folder {folder} as it does not match the expected format.")
        continue
    episode = int(fd_base)
    output_file_path = f"{output_dir}/{episode}.mcap"
    print(f"{output_file_path=}")
    mcap_writer = Writer(output_file_path)
    mcap_writer.start()
    flb_writer = McapFlatBuffersWriter()
    flb_writer.set_writer(mcap_writer)
    mcap_tool = McapTool(mcap_writer)
    all_schemas = set(FlatBuffersSchemas)
    all_schemas.remove(FlatBuffersSchemas.COMPRESSED_IMAGE)
    flb_writer.register_schemas(all_schemas)
    McapDataSampler.add_config_metadata(mcap_writer, config)
    # find all .mp4 files in the folder
    mp4_files = [
        f.path for f in os.scandir(folder) if f.is_file() and f.name.endswith(".mp4")
    ]
    print(f"{mp4_files=}")
    # add video attachments
    for mp4_file in mp4_files:
        with open(mp4_file, "rb") as f:
            name = f"/{os.path.basename(mp4_file).removesuffix('.mp4')}/color/image_raw"
            print(f"Adding video attachment: {name}")
            mcap_writer.add_attachment(
                time.time_ns(),
                time.time_ns(),
                name,
                MediaType.VIDEO_MP4,
                f.read(),
            )

    def to_topic(group: str, component: str) -> str:
        return f"/{group}/arm/pose/position"

    # load json dict
    groups = ["lead", "follow"]
    components = ["arm", "eef"]
    slices = [slice(0, 6), slice(6, 7)]

    path = f"{folder}/obs_action.json"
    if not os.path.exists(path):
        print(f"Skipping file {path} as it does not exist.")
        continue

    topic_mapping: Dict[str, dict] = {
        "jq": {
            "/follow/arm/joint_states/position": slice(0, 6),
            "/follow/eef/joint_states/position": slice(6, 7),
        },
        "jv": {
            "/follow/arm/joint_states/velocity": slice(0, 6),
            "/follow/eef/joint_states/velocity": slice(6, 7),
        },
        "tau": {
            "/follow/arm/joint_states/effort": slice(0, 6),
            "/follow/eef/joint_states/effort": slice(6, 7),
        },
        "eef_pos": "/follow/arm/pose/position",
        "end_force": {
            "/follow/arm/wrench/force": slice(0, 3),
            "/follow/arm/wrench/torque": slice(3, 6),
        },
        "act": "/lead/arm/pose/position",
    }

    topic_names = set()
    for key, value in topic_mapping.items():
        if isinstance(value, str):
            topic_mapping[key] = {value: None}
            topic_names.add(value)
        else:
            assert isinstance(value, dict), f"Value for key {key} must be a dict."
            topic_names.update(value.keys())

    with open(path) as f:
        act_obs: dict = json.load(f)
        print(f"{act_obs.keys()=}")
        obs: dict = act_obs.pop("obs", {})
        print(f"{obs.keys()=}")
        times = act_obs.pop("time", [])
        acts: list = act_obs.pop("act", [])
        not_mapping_keys = (act_obs.keys() | obs.keys()) - topic_mapping.keys()
        print(f"Not mapping keys: {not_mapping_keys}")
        topic_names.update(not_mapping_keys)
        print(f"All topic names: {topic_names}")
        # register joint state channels
        for topic in topic_names:
            flb_writer.register_channel(topic, FlatBuffersSchemas.FLOAT_ARRAY)
        # add joint states messages
        stamps_ns = []

        for stamp in times:
            stamps_ns.append(int(stamp * 1e9))
        mcap_writer.add_attachment(
            time.time_ns(),
            time.time_ns(),
        )
        mcap_tool.add_log_stamps_attachment(stamps_ns)

        for i, action in enumerate(acts):
            stamp_ns = stamps_ns[i]
            for topic, slc in topic_mapping["act"].items():
                if slc:
                    flb_writer.add_float_array(topic, action[slc], stamp_ns, stamp_ns)
                else:
                    flb_writer.add_float_array(topic, action, stamp_ns, stamp_ns)

        for key, values in obs.items():
            for topic, slc in topic_mapping.get(key, {key: None}).items():
                for i, value in enumerate(values):
                    stamp_ns = stamps_ns[i]
                    if slc:
                        flb_writer.add_float_array(
                            topic, value[slc], stamp_ns, stamp_ns
                        )
                    else:
                        flb_writer.add_float_array(topic, value, stamp_ns, stamp_ns)

    mcap_writer.finish()


print(
    f"Time taken: {time.perf_counter() - start:.3f} seconds of {len(folders)} folders"
)



from abc import abstractmethod
from typing import Any, Dict, Optional
from pathlib import Path
from pydantic import BaseModel, ConfigDict
from typing import Literal, Union, List
from airbot_data_collection.basis import ConfigurableBasis
from airbot_data_collection import __version__ as collector_version
from mcap_data_loader.utils.dict import CallableKeyMappingDict, MappingCall


class Subtask(BaseModel, frozen=True):
    skill: str
    """Skill template with placeholders like "pick {A} from {B}"."""
    description: str
    """English description of the subtask."""
    description_zh: str
    """Chinese description of the subtask."""


class TaskInfo(BaseModel, frozen=True):
    model_config = ConfigDict(extra="forbid")

    task_name: str = ""
    """Name of the task being performed. Used for identification and logging."""
    task_description: str = ""
    """Detailed description of the task in English."""
    task_description_zh: str = ""
    """Detailed description of the task in Chinese."""
    task_id: Union[str, int] = ""
    """Unique identifier for the task, used for tracking and management."""
    station: str = ""
    """Identifier for the station where the task is performed, useful for multi-station setups."""
    operator: str = ""
    """ID of the operator performing the task, useful for logging and accountability."""
    skill: Union[str, List[str]] = ""
    """Skill(s) being demonstrated or performed during the task."""
    object: Union[str, List[str]] = ""
    """Object(s) involved in the task."""
    scene: str = ""
    """Scene or environment description for the task."""
    subtasks: List[Subtask] = []
    """List of subtasks that compose the main task."""


class SaveType(BaseModel, frozen=True):
    model_config = ConfigDict(extra="forbid")

    color: Literal["raw", "jpeg", "h264"] = "h264"
    """Color image saving type."""
    depth: Literal["raw"] = "raw"
    """Depth image saving type."""


class Version(BaseModel, frozen=True):
    model_config = ConfigDict(extra="forbid")

    collector: str = collector_version
    """Version of the data collection codebase."""
    data_schema: str = "0.0.1"
    """Version of the data schema used for organizing and storing collected data."""


class DataSamplerConfig(BaseModel, frozen=True):
    model_config = ConfigDict(extra="forbid")

    version: Version = Version()
    """Version information for the data collection."""
    task_info: TaskInfo = TaskInfo()
    """Task information for the data collection."""
    save_type: SaveType = SaveType()
    """Data saving types for different modalities."""
    key_remap: MappingCall[str] = CallableKeyMappingDict()
    """Key remapping for data fields."""


class DataSampler(ConfigurableBasis):
    """Data sampler for sampling kinds of data.TODO: add a close method?"""

    @abstractmethod
    def compose_path(self, directory: Path, round: int) -> Path:
        """Compose the path to the data file. It will be called
        at starting sampling and removing. Before returning, file
        handler can be created to save data in `update` during sampling.
        Args:
            directory (Path): The directory where the data will be saved.
            round (int): The round number of the data.
        Returns:
            Path: The path to the data file.
        """

    def clear(self) -> None:
        """Clear the inner data buffer if any.
        Please be careful to avoid asynchronous saving
        exceptions caused by asynchronous clearing of data"""

    def update(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Process the data and return.
        Args:
            data (Dict[str, Any]): The data to be processed.
        Returns:
            Dict[str, Any]: The processed data, which
            will be append to the data buffer in the
            demonstration interface.
        """
        return data

    def remove(self, path: Path) -> Optional[bool]:
        """Remove the data from the given or last saved path.
        If the return value is None, the demonstrate
        interface will try to remove the path."""

    def set_info(self, info: Dict[str, Any]) -> None:
        """Set the info of the data collector.
        The info is a dict that contains the information
        of the data collector, such as the name, type, etc."""
        self._info = info

    @abstractmethod
    def save(self, path: Path, data: Any) -> bool:
        """Save the data to the given path.
        Args:
            path (Path): The path to the data file.
            data (Any): The data to be saved. If used in
            demonstration, the data are those stored in the
            data buffer of the demonstration interface.
        Returns:
            bool: True if the data was saved successfully, False otherwise.
        """


class MockDataSampler(DataSampler):
    """Mock data sampler for testing purpose."""

    def update(self, data) -> None:
        return None

    def save(self, path: Path, data: Any) -> bool:
        return True

    def remove(self, path: Path) -> bool:
        return True

    def compose_path(self, directory: Path, round: int) -> Path:
        return directory / f"mock_{round}.data"



import numpy as np
from typing import Any, Generic, TypeVar, final
from typing_extensions import Self
from collections.abc import Callable
from abc import abstractmethod
from pydantic import BaseModel, PositiveInt
from logging import getLogger
from airbot_data_collection.basis import ConfigurableBasis
from airbot_data_collection.common.environments.basis import (
    EnvironmentBasis,
    EnvironmentOutput,
)


T = TypeVar("T", bound=Callable[..., Any])


class WrapperBasis(ConfigurableBasis, Generic[T]):
    caller: Callable
    # the chained outputs of the wrappers, cleared on reset
    output_chain: list = []
    should_take_over_env: bool = False

    def on_configure(self) -> bool:
        return True

    @final
    def wrap(self, caller: T) -> Self:
        """Wrap the callable with additional functionality."""
        caller_type = self.__annotations__.get("caller", Callable)
        if not isinstance(caller, caller_type):
            raise TypeError(
                f"Expected callable of type {self.caller}, got {type(caller)}"
            )
        self._warmed_up = False
        self.caller = caller
        self.env = None
        return self

    @final
    def take_over_env(self, env: EnvironmentBasis):
        """Set the environment for the topmost wrapper.
        If the subclass needs to take over env, please
        set `self.should_take_over_env = True` in any of __init__ or warm_up.
        """
        if not self._warmed_up:
            raise RuntimeError("Please warm up first")
        if not self.should_take_over_env:
            raise RuntimeError("This wrapper has no need to take over the environment")
        if not isinstance(env, EnvironmentBasis):
            raise TypeError("The `env` must be an instance of EnvironmentBasis")
        self.env = env

    @final
    def warm_up(self, *args, **kwds) -> Any:
        """Warm up the wrapper by calling the caller once and process its output.
        The inputs are the same as the caller inputs."""
        if not hasattr(self, "caller"):
            raise RuntimeError("No caller, please wrap first")
        output = self.caller(*args, **kwds)
        if hasattr(output, "shape"):
            if len(output.shape) in (3, 5):
                self.output_type = type(output)
                self.output_type_name = type(output).__name__
                self.output_dtype = output.dtype
                if self.output_type_name == "ndarray":
                    backend = np
                    self.output_device = "cpu"
                elif self.output_type_name == "Tensor":
                    import torch as backend

                    self.output_device = output.device
                else:
                    raise TypeError(f"Unsupported output type: {self.output_type_name}")
                self.output_backend = backend
                self._warmed_up = True
                return self.on_warm_up(output)
            raise ValueError(
                "The shape of the caller output must be (B, T, C, H, W) or (B, T, D), "
                "i.e. (batch, time, channel, height, width) or (batch, time, dimension)."
                f"But got shape: {output.shape}"
            )
        raise TypeError("The caller output must have a shape")

    @final
    def reset(self):
        """Reset the internal state of the wrapper, if any."""
        if not self._warmed_up:
            raise RuntimeError("Please warm up first")
        if self.taken_over_env:
            self.env.reset()
            self.last_env_output = self.env.output()
        return self.on_reset()

    @abstractmethod
    def on_reset(self):
        """Hook when `reset` is called."""

    @abstractmethod
    def on_warm_up(self, output: Any):
        """Hook when `warm_up` is called. The `output` is the return value of the caller."""

    @abstractmethod
    def call(self, *args, **kwds) -> Any:
        """Call the wrapped callable and return the output."""

    @final
    def __call__(self, *args, **kwds):
        """Call the wrapped callable and update the output chain.
        If the env is not taken over, this method will return
        the output of the wrapped caller, and return the last output of env
        otherwise.
        """
        output = self.call(*args, **kwds)
        self.output_chain.append(output)
        if self.taken_over_env:
            self.env.input(output)
            return self.last_env_output
        return output

    @final
    def shutdown(self) -> bool:
        """Shutdown the wrapper and the environment."""
        if self.on_shutdown():
            if self.taken_over_env:
                return self.env.shutdown()
            return True
        return False

    def on_shutdown(self) -> bool:
        """Shutdown the wrapper, if any."""
        # NOTE: since most wrappers do not need to shutdown,
        # we provide a default implementation that does nothing
        return True

    @classmethod
    def get_logger(cls):
        return getLogger(cls.__name__)

    @classmethod
    def clear_output_chain(cls):
        cls.output_chain = []

    @staticmethod
    def calculate_size(data) -> float:
        type_name = type(data).__name__
        if type_name == "Tensor":
            size_mb = data.numel() * data.element_size() / (1024**2)
        elif type_name == "ndarray":
            size_mb = data.nbytes / (1024**2)
        else:
            raise TypeError(f"Unsupported type name: {type_name}")
        return size_mb

    @property
    def taken_over_env(self) -> bool:
        """Whether the wrapper takes over the environment"""
        return self.env is not None


class ForwardingWrapper(WrapperBasis):
    """A wrapper that just forwards the call to the caller"""

    def on_warm_up(self, output: Any):
        return output

    def on_reset(self):
        """Do nothing"""

    def call(self, *args, **kwds) -> Any:
        return self.caller(*args, **kwds)


class TakeOverEnvWrapper(WrapperBasis):
    """A wrapper that just takes over the environment"""

    should_take_over_env = True

    def on_warm_up(self, output: Any):
        return output

    def on_reset(self):
        """Do nothing"""

    def call(self) -> EnvironmentOutput:
        """Call the environment, using the output
        observation as the caller input and returning
        the environment output
        """
        env_output = self.env.output()
        call_output = self.caller(env_output.observation)
        self.env.input(call_output)
        self.last_env_output = env_output
        return env_output

    def __call__(self, *args, **kwds) -> EnvironmentOutput:
        return super().__call__(*args, **kwds)


class NormalizationConfig(BaseModel):
    mean: float = 0.0
    std: float = 0.0
    min_std: float = 1e-8

    def model_post_init(self, context):
        self.std = self.std or self.min_std


class NormalizerConfig(BaseModel):
    """Configuration to normalize the input and denormalize the output"""

    input: NormalizationConfig = NormalizationConfig()
    output: NormalizationConfig = NormalizationConfig()
    # if input data is a dict and input_key is not empty, normalize the data[input_key]
    input_key: str = ""


class Normalizer(WrapperBasis):
    config: NormalizerConfig

    def normalize_input(self, data: Any) -> Any:
        if isinstance(data, dict):
            key = self.config.input_key
            if not key:
                raise ValueError("data is a dict, but key is empty")
            data[key] = self._normalize(data[key])
            return data
        return self._normalize(data)

    def _normalize(self, data: Any) -> Any:
        return (data - self.config.input.mean) / self.config.input.std

    def denormalize_output(self, data: Any) -> Any:
        return data * self.config.output.std + self.config.output.mean

    def on_warm_up(self, output: Any):
        """Do nothing"""

    def on_reset(self):
        """Do nothing"""

    def call(self, *args, **kwds):
        return self.denormalize_output(self.caller(self.normalize_input(*args, **kwds)))


class FrequencyReductionCallConfig(BaseModel):
    # 1 means call at each step
    period: PositiveInt = 1


class FrequencyReductionCall(WrapperBasis):
    config: FrequencyReductionCallConfig

    def on_warm_up(self, output: Any):
        horizon = output.shape[1]
        period = self.config.period
        if horizon < period:
            raise ValueError(
                f"output horizon: {horizon} can not be shorter than period: {period}"
            )
        return output

    def on_reset(self):
        self._t = 0

    def call(self, *args, **kwds):
        target_t = self._t % self.config.period
        if target_t == 0:
            self._outputs = self.caller(*args, **kwds)
        self._t += 1
        return self._outputs[:, target_t]



import subprocess
import platform
from typing import Dict, Any, Optional
import re
import os


class SystemInfo:
    @classmethod
    def get_product(cls, with_sudo: bool = False) -> Dict[str, Any]:
        if with_sudo:
            return cls._get_product_dmidecode()
        else:
            try:
                return {
                    "product_name": subprocess.check_output(
                        "cat /sys/devices/virtual/dmi/id/product_name",
                        shell=True,
                        text=True,
                    ).strip(),
                }
            except subprocess.CalledProcessError:
                # 如果在容器中无法访问DMI信息，返回有限的信息
                return {"product_name": "Unknown (container environment)"}

    @staticmethod
    def _get_product_dmidecode():
        try:
            result = subprocess.check_output(
                "sudo dmidecode -t system", shell=True, text=True
            )
            info = {}
            for line in result.splitlines():
                if "Manufacturer:" in line:
                    info["manufacturer"] = line.strip().split(":", 1)[1].strip()
                elif "Product Name:" in line:
                    info["product_name"] = line.strip().split(":", 1)[1].strip()
                elif "Version:" in line:
                    info["version"] = line.strip().split(":", 1)[1].strip()
                elif "Serial Number:" in line:
                    info["serial_number"] = line.strip().split(":", 1)[1].strip()
                elif "UUID:" in line:
                    info["uuid"] = line.strip().split(":", 1)[1].strip()
            return info
        except Exception as e:
            print(f"Error accessing dmidecode: {e}")
            return {"product_name": "Unknown (container environment)"}

    @staticmethod
    def get_cpu():
        info = {
            "model_name": "",
            "cores_physical": "0",
            "cores_logical": "0",
        }

        try:
            with open("/proc/cpuinfo") as f:
                cpuinfo = f.read()

            for line in cpuinfo.splitlines():
                if line.startswith("model name"):
                    info["model_name"] = line.split(":", 1)[1].strip()
                elif line.startswith("siblings"):
                    info["cores_logical"] = line.split(":", 1)[1].strip()
                elif line.startswith("cpu cores"):
                    info["cores_physical"] = line.split(":", 1)[1].strip()

            # 如果未能获取物理核心数，尝试使用nproc命令
            if info["cores_physical"] == "0":
                try:
                    info["cores_physical"] = subprocess.check_output(
                        "nproc", shell=True, text=True
                    ).strip()
                except subprocess.CalledProcessError:
                    pass
        except Exception as e:
            print(f"Error reading CPU info: {e}")

        return info

    @staticmethod
    def get_memory():
        mem_info = {
            "mem_total": "Unknown",
            "mem_free": "Unknown",
            "swap_total": "0 kB",
            "swap_free": "0 kB",
        }

        try:
            with open("/proc/meminfo") as f:
                meminfo = f.read()

            for line in meminfo.splitlines():
                if "MemTotal" in line:
                    mem_info["mem_total"] = line.split(":", 1)[1].strip()
                elif "MemFree" in line:
                    mem_info["mem_free"] = line.split(":", 1)[1].strip()
                elif "SwapTotal" in line:
                    mem_info["swap_total"] = line.split(":", 1)[1].strip()
                elif "SwapFree" in line:
                    mem_info["swap_free"] = line.split(":", 1)[1].strip()
        except Exception as e:
            print(f"Error reading memory info: {e}")

        return mem_info

    @staticmethod
    def get_gpu():
        """
        读取lspci命令输出并解析VGA设备信息为结构化字典，适应容器环境

        Returns:
            dict: 包含VGA设备信息的字典，键为设备编号，值为设备属性字典
        """
        try:
            # 首先检查lspci命令是否可用
            subprocess.check_call("which lspci > /dev/null 2>&1", shell=True)

            output = subprocess.check_output(
                "lspci | grep -i 'vga\\|3d\\|display'", shell=True, text=True
            )
            if not output.strip():
                return {
                    "info": "No GPU detected or information not available in container"
                }

            lines = output.strip().split("\n")

            devices = {}
            for i, line in enumerate(lines, 1):
                # 正则表达式匹配PCI地址、设备类型、厂商、型号和修订版本
                match = re.match(
                    r"([0-9a-f:.]+)\s+([^:]+):\s+([^[]+)\[([^]]+)\](?:\s+\(rev\s+([0-9a-f]+)\))?",
                    line,
                )

                if match:
                    pci_address, device_type, vendor, model, revision = match.groups()

                    # 清理字符串
                    pci_address = pci_address.strip()
                    device_type = device_type.strip()
                    vendor = vendor.strip()
                    model = model.strip()
                    revision = revision.strip() if revision else ""

                    devices[f"device{i}"] = {
                        "pci_address": pci_address,
                        "device_type": device_type,
                        "vendor": vendor,
                        "model": model,
                        "revision": revision,
                    }
                else:
                    # 尝试另一种格式的匹配
                    match = re.match(
                        r"([0-9a-f:.]+)\s+([^:]+):\s+([^(]+)(?:\s+\(rev\s+([0-9a-f]+)\))?",
                        line,
                    )
                    if match:
                        pci_address, device_type, vendor_model, revision = (
                            match.groups()
                        )

                        # 清理字符串
                        pci_address = pci_address.strip()
                        device_type = device_type.strip()
                        vendor_model = vendor_model.strip()
                        revision = revision.strip() if revision else ""

                        # 尝试分离厂商和型号
                        parts = vendor_model.split(" ", 1)
                        vendor = parts[0]
                        model = parts[1] if len(parts) > 1 else ""

                        devices[f"device{i}"] = {
                            "pci_address": pci_address,
                            "device_type": device_type,
                            "vendor": vendor,
                            "model": model,
                            "revision": revision,
                        }

            return devices

        except subprocess.CalledProcessError:
            # 尝试使用nvidia-smi获取NVIDIA GPU信息
            try:
                subprocess.check_call("which nvidia-smi > /dev/null 2>&1", shell=True)
                output = subprocess.check_output(
                    "nvidia-smi --query-gpu=name,driver_version,memory.total --format=csv,noheader",
                    shell=True,
                    text=True,
                )

                devices = {}
                for i, line in enumerate(output.strip().split("\n"), 1):
                    parts = line.split(", ")
                    if len(parts) >= 3:
                        devices[f"device{i}"] = {
                            "vendor": "NVIDIA",
                            "model": parts[0].strip(),
                            "driver_version": parts[1].strip(),
                            "memory": parts[2].strip(),
                        }
                return devices
            except subprocess.CalledProcessError:
                return {
                    "info": "GPU information not available in container environment"
                }
        except Exception as e:
            print(f"Error getting GPU info: {e}")
            return {"info": f"Error retrieving GPU information: {str(e)}"}

    @staticmethod
    def is_in_docker() -> bool:
        """检测是否在Docker容器中运行"""
        return (
            os.path.exists("/.dockerenv")
            or os.path.isfile("/proc/self/cgroup")
            and any("docker" in line for line in open("/proc/self/cgroup"))
        )

    @staticmethod
    def get_platform() -> Dict[str, Any]:
        platform_info = platform.uname()._asdict()

        # 添加容器信息
        if SystemInfo.is_in_docker():
            platform_info["environment"] = "Docker container"

            # 尝试获取容器ID
            try:
                with open("/proc/self/cgroup") as f:
                    for line in f:
                        if "docker" in line:
                            container_id = line.split("/")[-1].strip()
                            platform_info["container_id"] = container_id
                            break
            except Exception:
                pass

        return platform_info

    @classmethod
    def all_info(cls, with_sudo: bool = False) -> Dict[str, Any]:
        info = {
            "product": cls.get_product(with_sudo),
            "cpu": cls.get_cpu(),
            "memory": cls.get_memory(),
            "gpu": cls.get_gpu(),
            "platform": cls.get_platform(),
        }

        # 添加容器环境标记
        if cls.is_in_docker():
            info["container_environment"] = True

        return info


if __name__ == "__main__":
    from pprint import pprint

    system_info = SystemInfo.all_info()
    pprint(system_info)



from multiprocessing.managers import SharedMemoryManager
from multiprocessing import get_context, current_process
from multiprocessing.connection import Connection
from pydantic import BaseModel, ConfigDict, ImportString, validate_call
from typing import Union, Dict, Type, Optional
from airbot_data_collection.common.systems.basis import Sensor, System
from airbot_data_collection.basis import ConcurrentMode
from airbot_data_collection.common.utils.event_rpc import (
    EventRpcManager,
    EventRpcServer,
)
from airbot_data_collection.utils import init_logging
from airbot_data_collection.common.utils.shareable_numpy import ShareableNumpy
from airbot_data_collection.common.utils.shareable_value import ShareableValue
from numpy import uint64
from setproctitle import setproctitle

InterfaceType = Union[Sensor, System]


class ConcurrentWrapperConfig(BaseModel):
    """The config for the concurrent wrapper."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    interface: InterfaceType
    """the interface instance to be wrapped"""
    concurrent: ConcurrentMode = ConcurrentMode.process
    """the concurrent mode"""


class SensorConcurrentWrapper(Sensor):
    """A wrapper for sensors to run in a separate thread or process."""

    def __init__(self, config: ConcurrentWrapperConfig):
        self._concurrent = config.concurrent
        self._interface = config.interface

    # def __call__(self, interface: InterfaceType):
    #     self._interface = interface
    #     return self

    def on_configure(self) -> bool:
        # TODO: use protocol instead of inheriting Sensor?
        if self._concurrent is not ConcurrentMode.process:
            raise NotImplementedError(
                "Only process mode is supported for SensorConcurrentWrapper."
            )
        self._rpc = EventRpcManager(EventRpcManager.get_args(self._concurrent))
        spawn_ctx = get_context("spawn")
        parent, child = spawn_ctx.Pipe()
        self._smm = SharedMemoryManager(ctx=spawn_ctx)
        self._smm.start()
        self._concurrent = EventRpcManager.get_concurrent_cls(self._concurrent)(
            target=self._concurrent_loop,
            args=(self._interface, child, self._rpc.server),
            name=f"{self.__class__.__name__}Concurrent",
        )
        self._concurrent.start()
        if parent.poll(5.0):
            if parent.recv():
                self.get_logger().info("Receiving info and first observation...")
                if parent.poll(5.0):
                    self._info, obs = parent.recv()
                    self._obs: Dict[
                        str, Dict[str, Union[ShareableNumpy, ShareableValue]]
                    ] = {}
                    obs: dict
                    for key, value in obs.items():
                        self._obs[key] = {
                            "t": ShareableValue.from_value(
                                value["t"], uint64, smm=self._smm
                            ),
                            "data": ShareableNumpy.from_array(
                                value["data"], smm=self._smm
                            ),
                        }
                    self.get_logger().info("Sending shm observation...")
                    parent.send(self._obs)
                    parent.close()
                    return True
            else:
                self.get_logger().error("Failed to configure the interface.")
        else:
            self.get_logger().error("Timeout waiting for configure response.")
        return False

    @staticmethod
    def _concurrent_loop(
        interface: Sensor, conn: Connection, rpc_server: EventRpcServer
    ):
        init_logging()
        logger = interface.get_logger()
        logger.info("Configuring the interface...")
        setproctitle(f"{current_process().name}:{logger.name}")
        conn.send(interface.configure())
        conn.send((interface.get_info(), interface.capture_observation(5.0)))
        if not conn.poll(5.0):
            logger.error("Timeout waiting for shm observation.")
            return
        shm_obs = conn.recv()
        conn.close()
        while rpc_server.wait():
            for key, value in interface.capture_observation(5.0).items():
                shm_obs[key]["t"].value = value["t"]
                shm_obs[key]["data"][:] = value["data"]
            rpc_server.respond()
        logger.info("Shutting down the interface...")
        interface.shutdown()

    def _get_obs(self):
        return {
            key: {"t": value["t"].value, "data": value["data"].array}
            for key, value in self._obs.items()
        }

    def capture_observation(self, timeout: Optional[float] = None):
        if self._rpc.client.request(timeout=timeout):
            if timeout is not None and timeout <= 0:
                return None
            return self._get_obs()
        raise TimeoutError("Timeout waiting for observation.")

    def result(self, timeout: Optional[float] = None):
        if timeout is None or timeout > 0:
            if self._rpc.client.wait(timeout=timeout):
                return self._get_obs()
            raise TimeoutError("Timeout waiting for observation.")
        raise ValueError("Timeout must be None or positive.")

    def get_info(self):
        return self._info

    def shutdown(self):
        self._rpc.shutdown()
        self._concurrent.join(timeout=5.0)
        self._smm.shutdown()
        if self._concurrent.is_alive():
            self.get_logger().warning("Concurrent process did not terminate in time.")
        return True


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def concurrent_wrapper(
    interface_cls: Union[Type[InterfaceType], ImportString[Type[InterfaceType]]],
    config_cls: Optional[Type[BaseModel]] = None,
):
    config_cls = config_cls or interface_cls.resolve_config_type(interface_cls)
    field_name = "concurrent_wrapped"
    if config_cls is None:
        ConfigWithConcurrent = ConcurrentWrapperConfig
    else:
        if field_name not in config_cls.model_fields:

            class ConfigWithConcurrent(config_cls):
                concurrent_wrapped: ConcurrentMode = ConcurrentMode.process
                """the wrapped concurrent mode"""
        else:
            # TODO: use logging instead of raising error?
            raise ValueError(
                f"The config class {config_cls.__name__} already has a field named '{field_name}'."
            )
            ConfigWithConcurrent = config_cls

    class ConcurrentWrappedClass(SensorConcurrentWrapper):
        """A concurrent wrapper class for the given interface class."""

        def __init__(self, config: ConfigWithConcurrent):
            self._concurrent = config.concurrent_wrapped
            """
            Interface configurations may include a `concurrent` parameter, typically used in conjunction with a `blocking` parameter to implement asynchronous, non-blocking reads. The former is a necessary condition for non-blocking, but asynchronous does not equal non-blocking. Blocking essentially ensures that data timestamps are not duplicated, while asynchronous operation determines how timestamps are updated. Even in asynchronous mode, blocking can still occur if the data timestamp has not been updated.

            When using asynchronous mode directly with an interface, using a child process is reasonable to maximize asynchronous loop speed and avoid interference from other threads in the current process. However, when combined with a wrapper, if the wrapper itself is an independent process, the child process mode within the interface becomes unnecessary. Even so, forced exceptions should not be added here, as there may be other thread contention or other considerations within the interface. At most, a warning message might be issued. Similarly, other combinations of inner and outer threads may also have certain considerations. For example, if both the wrapper and the inner thread are threads, it seems somewhat redundant, but if the interface involves I/O-intensive operations, the inner thread is sometimes necessary.

            In summary, adding additional checks and logs is not currently being considered.
            """
            # TODO:
            # inner_con = self.__get_inner_con(config)
            # if inner_con is self._concurrent:
            #     self.get_logger().warning(
            #         f"The inner concurrent {inner_con} is ignored in favor of the wrapped one {self._concurrent}."
            #     )
            # elif (
            #     self._concurrent is ConcurrentMode.thread
            #     and inner_con is ConcurrentMode.process
            # ):
            #     self.get_logger().warning(
            #         f"The inner concurrent {inner_con} is more powerful than the wrapped one {self._concurrent}."
            #     )
            # NOTE: use the original config to avoid errors caused by local
            # classes defined inside functions being unable to be pickled
            config_cls_dict = dict(config)
            config_cls_dict.pop(field_name)
            self._interface = interface_cls(config_cls(**config_cls_dict))

        # @classmethod
        # def __get_inner_con(cls, config):
        #     return getattr(config, "concurrent", None)

        # def __new__(cls, config: ConfigWithConcurrent):
        #     if config.concurrent_wrapped is ConcurrentMode.thread:
        #         with ForceSetAttr(config):
        #             if cls.__get_inner_con(config) is not None:
        #                 cls.get_logger().warning(f"Use the inner concurrent ")
        #                 config.concurrent = ConcurrentMode.thread
        #                 return interface_cls(config)
        #     return super().__new__(cls, config)

    return ConcurrentWrappedClass, ConfigWithConcurrent


def concurrent_instantiate(
    interface_cls, config_cls=None, config_kwargs: dict = None, **kwargs
) -> SensorConcurrentWrapper:
    cls, cfg = concurrent_wrapper(interface_cls, config_cls)
    config_kwargs = config_kwargs or {}
    config_kwargs.update(kwargs)
    return cls(cfg(**config_kwargs))


if __name__ == "__main__":
    from airbot_data_collection.common.devices.cameras.mock import MockCamera
    import cv2
    import time

    init_logging()

    cls = "airbot_data_collection.common.devices.cameras.mock.MockCamera"
    cls_wrapped, cfg_wrapped = concurrent_wrapper(cls)
    cls = MockCamera
    cls_wrapped, cfg_wrapped = concurrent_wrapper(cls)
    # con_mock_cam = cls_wrapped(
    #     cfg_wrapped(random=True, concurrent_wrapped=ConcurrentMode.process)
    # )
    con_mock_cam = concurrent_instantiate(cls)
    assert con_mock_cam.configure()
    con_mock_cam.get_logger().info("Successfully configured")
    for i in range(10):
        start = time.perf_counter()
        obs = con_mock_cam.capture_observation()
        print(f"{(time.perf_counter() - start) * 1000:.3} ms")
        for key, value in obs.items():
            print(key, value["t"])
            cv2.imshow(key, value["data"])
            cv2.waitKey(1)

    cv2.destroyAllWindows()
    con_mock_cam.shutdown()



from airbot_data_collection.common.systems.grouped import (
    GroupedComponentsSystemConfig,
    GroupedComponentsSystem,
    GroupsSendActionConfig,
)
from airbot_data_collection.basis import DictDataStamped, ForceSetAttr
from pydantic import field_validator, ValidationInfo
from mcap_data_loader.datasets.dataset import RealTimeDatasetABC


class GroupedSystemDataSourceConfig(GroupedComponentsSystemConfig):
    """Configuration for Grouped Components System Data Source"""

    reset_action: GroupsSendActionConfig = GroupsSendActionConfig()
    """Action to reset the grouped components system"""

    @field_validator("reset_action")
    def validate_reset_action(cls, v: GroupsSendActionConfig, info: ValidationInfo):
        with ForceSetAttr(v):
            if v.action_values and not v.groups:
                v.groups = list(dict.fromkeys(info.data["components"].groups))
            v.model_post_init(None)
            return v


class GroupedSystemDataSource(RealTimeDatasetABC[DictDataStamped]):
    """Grouped Components System Data Source"""

    def __init__(self, config: GroupedSystemDataSourceConfig):
        self.config = config
        self.interface = GroupedComponentsSystem(config)
        if not self.interface.configure():
            raise RuntimeError("Failed to configure GroupedComponentsSystem")

    def reset(self):
        return self.interface.send_action(self.config.reset_action)

    def write(self, input: GroupsSendActionConfig):
        return self.interface.send_action(input)

    def read(self):
        return self.interface.capture_observation()

    def close(self):
        return self.interface.shutdown()


if __name__ == "__main__":
    from airbot_data_collection.utils import init_logging
    from airbot_data_collection.common.systems.basis import SystemMode
    from airbot_data_collection.common.systems.grouped import (
        SystemSensorComponentGroupsConfig,
        AutoControlConfig,
    )
    from pprint import pprint
    from airbot_data_collection.common.devices.cameras.mock import MockCamera
    from airbot_data_collection.common.systems.mock import MockSystem
    import time

    init_logging()

    reset_action = GroupsSendActionConfig(
        groups=["/"], action_values=[[0.0] * 6], modes=[SystemMode.RESETTING]
    )
    system = MockSystem()
    camera = MockCamera()
    config = GroupedSystemDataSourceConfig(
        components=SystemSensorComponentGroupsConfig(
            groups=["/", "/"],
            names=["lead_arm", "left_camera"],
            roles=["l", "o"],
            instances=[system, camera],
        ),
        reset_action=reset_action,
        auto_control=AutoControlConfig(groups=[]),
    )
    pprint(config.model_dump())
    live_data = GroupedSystemDataSource(config)
    # assert live_data.configure()
    live_data.reset()
    print(live_data.read().keys())
    reset_action.modes = [SystemMode.SAMPLING] * len(reset_action.modes)
    live_data.write(reset_action)

    for i, data in enumerate(live_data):
        for key, value in data.items():
            print(f"{key}: {value['t']}")
        time.sleep(0.5)
        if i >= 5:
            break
    live_data.close()



from pydantic import BaseModel, NonNegativeFloat, NonNegativeInt, ConfigDict
from typing import Any, Optional
from airbot_data_collection.common.systems.basis import System, SystemMode
from more_itertools import consume, seekable
from mcap_data_loader.datasets.dataset import IterableDatasetABC


class DataPlayerConfig(BaseModel, frozen=True):
    """Configuration for the data player."""

    model_config = ConfigDict(extra="forbid")

    source: Any
    loop: bool = False
    rate: NonNegativeFloat = 1.0
    start_time: NonNegativeInt = 0
    end_time: NonNegativeInt = 0
    cache: bool = False


class IterablePlayer(System):
    """A system that plays back data from a stream."""

    config: DataPlayerConfig

    def on_configure(self) -> bool:
        interface: IterableDatasetABC = self.interface
        if self.config.cache:
            self._stream = seekable(interface)
        else:
            self._stream = iter(interface)
        return True

    def _create_interface(self, class_type: IterableDatasetABC):
        return class_type(self.config.source)

    def send_action(self, action: int):
        """Set the stream position to the action index."""
        if self.config.cache:
            self._stream.seek(action)
        else:
            self._stream = iter(self.interface)
            consume(self._stream, action)

    def on_switch_mode(self, mode: SystemMode):
        if mode == SystemMode.RESETTING:
            self.send_action(0)
        return True

    def capture_observation(self, timeout: Optional[float] = None) -> Any:
        return next(self._stream, None)

    def get_info(self) -> dict:
        return {}

    def shutdown(self) -> bool:
        return True



import logging
from typing import Optional, List, Dict
from airbot_data_collection.common.devices.vr.quest import (
    VRQuest,
    VRQuestConfig,
    VRControllerEvent,
    EventValueMode,
)
from airbot_data_collection.basis import Bcolors
from pprint import pformat
from functools import partial
from abc import ABC, abstractmethod
from pydantic import BaseModel


class InputController(ABC):
    """Base class for input controllers that generate motion deltas."""

    def __init__(self):
        """
        Initialize the controller.

        Args:
            step_size: Base movement step size in meters / rad
        """
        self.running = True
        self.episode_end_status = None  # None, "success", or "failure"
        self.intervention_flag = False
        self.open_gripper_command = False
        self.close_gripper_command = False

    def start(self):
        """Start the controller and initialize resources."""
        pass

    def stop(self):
        """Stop the controller and release resources."""
        pass

    @abstractmethod
    def get_deltas(self):
        """Get the current movement deltas in standard units."""
        raise NotImplementedError

    def should_quit(self) -> bool:
        """Return True if the user has requested to quit."""
        return not self.running

    def update(self):
        """Update controller state - call this once per frame."""
        pass

    def __enter__(self):
        """Support for use in 'with' statements."""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Ensure resources are released when exiting 'with' block."""
        self.stop()

    def get_episode_end_status(self) -> Optional[str]:
        """
        Get the current episode end status.

        Returns:
            None if episode should continue, "success" or "failure" otherwise
        """
        status = self.episode_end_status
        self.episode_end_status = None  # Reset after reading
        return status

    def should_intervene(self) -> bool:
        """Return True if intervention flag was set."""
        return self.intervention_flag

    def gripper_command(self) -> Optional[str]:
        """Return the current gripper command."""
        if self.open_gripper_command == self.close_gripper_command:
            return "stay"
        elif self.open_gripper_command:
            return "open"
        elif self.close_gripper_command:
            return "close"

    def get_logger(self):
        return logging.getLogger(self.__class__.__name__)


class EventConfig(BaseModel):
    zero_info: int
    success: int
    failure: int
    rerecord_episode: int
    intervention: int
    shutdown: int
    left_eef: int
    right_eef: int


class TeleopConfig(BaseModel):
    """
    Configuration for teleoperation controllers.
    This can be extended to include more parameters as needed.
    """

    event_config: EventConfig
    pos: Optional[List[str]] = None


class VRQuestController(InputController):
    """Generate motion deltas from vr meta quest3 input"""

    def __init__(self, config: TeleopConfig):
        super().__init__()
        self.intervention_flag = False
        event_config = config.event_config
        self._event_config = event_config
        self._pos = config.pos if config.pos is not None else ["left", "right"]
        self._end_status = {"success", "failure", "rerecord_episode"}
        self._event_to_end_status: Dict[int, str] = {
            getattr(event_config, field): field for field in self._end_status
        }
        self._init_vr()
        self._show_insructions()

    def _show_insructions(self):
        instructions = {status: status for status in self._end_status} | {
            "intervention": "start/stop intervention",
            "left_eef": "open/close left_eef",
            "right_eef": "open/close right_eef",
            "zero_info": "set current pose as zero for rela-control and start/stop intervention",
            "shutdown": "stop (exit/quit) the teleoperation",
        }
        event_type = self._vr._get_event_type()
        self.get_logger().info(
            Bcolors.blue(
                "\n"
                + pformat(
                    {
                        event_type(event).name: instructions[field]
                        for field, event in self._event_config.model_dump().items()
                    }
                )
            )
        )

    def _init_vr(self):
        self._vr = VRQuest(
            VRQuestConfig(
                zero_info={comp: self._event_config.zero_info for comp in self._pos}
            )
        )

    def start(self):
        """Initialize vr quest."""
        if not self._vr.configure():
            self.running = False
            return
        for event, status in self._event_to_end_status.items():
            self._vr.register_event_callback(
                event,
                partial(self._set_episode_end_status, status),
                EventValueMode.LEAVE_ZERO,
            )
            self._vr.register_event_callback(
                event,
                self._clear_episode_end_status,
                EventValueMode.ENTER_ZERO,
            )
        self.get_logger().info("Started.")

    def stop(self):
        """Clean up pygame resources."""
        self._vr.shutdown()
        self.get_logger().info("Stopped.")

    def update(self):
        """Process pygame events to get fresh gamepad readings."""
        control_data = self._vr.get_control_data()
        self.update_eef(control_data)
        if control_data[self._event_config.intervention]:
            self.intervention_flag = True
        else:
            self.intervention_flag = False
        if control_data[self._event_config.shutdown]:
            self.running = False

    def update_eef(self, control_data=None):
        """Update the end-effector command based on gamepad input."""
        if control_data is None:
            control_data = self._vr.get_control_data()
        # TODO: support both left and right eef
        eef = control_data[self._event_config.right_eef]
        if eef < 0:
            self.close_gripper_command = True
        elif eef > 0:
            self.open_gripper_command = True
        else:
            self.close_gripper_command = False
            self.open_gripper_command = False

    def get_deltas(self, pos: str) -> List[float]:
        """Get the current movement deltas from vr state."""
        return self._vr.get_rela_info_data(pos)

    def _set_episode_end_status(self, status: str, data: float):
        """Set the episode end status based on gamepad input."""
        self.episode_end_status = status
        # self.get_logger().info(f"Episode end status set to: {status}")

    def _clear_episode_end_status(self, data: float):
        """Clear the episode end status after reading it."""
        self.episode_end_status = None
        # self.get_logger().info("Episode end status cleared.")


def main(controller: VRQuestController):
    import time
    from airbot_data_collection.utils import init_logging
    from airbot_data_collection.common.utils.transformations import (
        euler_from_quaternion,
        quaternion_from_euler,
    )
    from airbot_data_collection.common.utils.coordinate import CoordinateTools
    import numpy as np

    np.set_printoptions(precision=3)

    init_logging(logging.INFO)

    controller.start()

    try:
        while not controller.should_quit():
            controller.update()
            if status := controller.get_episode_end_status():
                controller.get_logger().info(f"Episode ended with status: {status}")
            if controller.should_intervene():
                right_deltas = controller.get_deltas("right")
                left_deltas = controller.get_deltas("left")
                eef = controller.gripper_command()
                # controller.get_logger().info(f"Current deltas: {deltas}, eef: {eef}")
                # controller.get_logger().info(f"Delta euler angles: {euler_from_quaternion(deltas[3:7])}")
                # controller.get_logger().info(f"Delta position: {pose[0]}, eef: {eef}")
                # controller.get_logger().info(f"Delta euler angles: {euler_from_quaternion(pose[1])}")
                all_abs_data = controller._vr.get_info_data()
                right_abs = all_abs_data["right"]
                left_abs = all_abs_data["left"]
                right_rela_left = CoordinateTools.to_world_coordinate(
                    (right_deltas[:3], right_deltas[3:7]), (left_abs[:3], left_abs[3:7])
                )
                controller._vr._tf_pub.broadcast_tf(
                    right_rela_left[0],
                    quaternion_from_euler(*right_rela_left[1]),
                    "right_rela_left",
                )
                # controller.get_logger().info(
                #     f"Absolute position: {np.array(right_abs[0:3])}"
                # )
                # controller.get_logger().info(
                #     f"Absolute euler: {np.array(euler_from_quaternion(right_abs[3:7]))}"
                # )
                controller.get_logger().info(f"eef: {eef}")
                # controller.get_logger().info(f"Absolute position: {np.array(abs_data[0])}")
                # controller.get_logger().info(f"Absolute euler: {np.array(euler_from_quaternion(abs_data[1]))}")
            time.sleep(0.1)  # Simulate frame delay
    finally:
        controller.stop()


if __name__ == "__main__":
    main(
        VRQuestController(
            TeleopConfig(
                event_config=EventConfig(
                    zero_info=VRControllerEvent.RIGHT_GRIP,
                    success=VRControllerEvent.Y,
                    failure=VRControllerEvent.X,
                    rerecord_episode=VRControllerEvent.LEFT_GRIP,
                    intervention=VRControllerEvent.RIGHT_GRIP,
                    shutdown=VRControllerEvent.B,
                    left_eef=VRControllerEvent.LEFT_STICK_V,
                    right_eef=VRControllerEvent.RIGHT_STICK_V,
                )
            )
        )
    )



from concurrent.futures import (
    Executor,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    Future,
    as_completed,
)
from logging import getLogger
from typing import Any, List, Dict, Union
from send2trash import send2trash
from airbot_data_collection.common import DataSampler, MockDataSampler, Visualizer
from airbot_data_collection.demonstrate.configs import (
    ConcurrentMode,
    DemonstrateAction,
    DemonstrateConfig,
)
from airbot_data_collection.demonstrate.basis import SampleInfo
from airbot_data_collection.basis import Bcolors
from airbot_data_collection.utils import get_items_by_ext, zip
from airbot_data_collection.common.utils.system_info import SystemInfo
from airbot_data_collection.common.utils.progress import ProgressBar
from airbot_data_collection.common.demonstrators.basis import Demonstrator
from airbot_data_collection.state_machine.basis import CallbackEventType
from collections import defaultdict
from functools import partial
from pathlib import Path
from tqdm import tqdm
import time
import shutil


class DemonstrateInterface:
    def __init__(self, config: DemonstrateConfig):
        self._config = config
        # init sampler, visualizers and demonstrator
        self._sampler = (
            config.sampler.instance
            if config.sampler.instance is not None
            else MockDataSampler()
        )
        self._visualizers = config.visualizers.instance_dict
        self._demonstrator = config.demonstrator.instance
        # init sample info
        start_round = self._config.sample_limit.start_round
        if start_round < 0:
            # detect the number of files in the directory
            ds = self._config.dataset
            start_round = (
                len(get_items_by_ext(ds.absolute_directory, ds.file_extension))
                + start_round
                + 1
            )
            self._sample_limit = self._config.sample_limit.model_copy(
                update={"start_round": start_round}
            )
        else:
            self._sample_limit = self._config.sample_limit
        self._sample_info = SampleInfo(round=start_round)
        # init concurrent actions
        concur = self._config.concurrent
        self._action_executors: Dict[DemonstrateAction, Executor] = {}
        mode2executor = {
            ConcurrentMode.thread: ThreadPoolExecutor,
            ConcurrentMode.process: ProcessPoolExecutor,
        }
        for action, mode, max_workers in zip(
            concur.actions, concur.modes, concur.max_workers
        ):
            args = (action.name,) if mode is ConcurrentMode.thread else ()
            self._action_executors[action] = mode2executor[mode](max_workers, *args)
        self._action_futures: Dict[DemonstrateAction, List[Future]] = defaultdict(list)
        # store current round data
        self._round_data = defaultdict(list)
        self._metrics = defaultdict(dict)
        self._register_fsm_callbacks()

    def get_logger(self):
        """
        Get the logger for the demonstration.
        """
        return getLogger(self.__class__.__name__)

    def configure(self) -> bool:
        """
        Configure all the components.
        """
        # set info before configuring the sampler
        # so that the sampler can use it for configuring
        names = list(self._visualizers.keys())
        components = list(self._visualizers.values())
        types = ["visualizer"] * len(self._visualizers)
        if self._configure_components(names, components, types):
            self._sampler.set_info(
                self._demonstrator.get_info() | {"system": SystemInfo.all_info()}
            )
            if self._configure_components(["sampler"], [self._sampler], ["sampler"]):
                return True
        return False

    def _configure_components(
        self,
        names: List[str],
        components: List[Union[Visualizer, DataSampler]],
        types: List[str],
    ) -> bool:
        for name, component, tp in zip(names, components, types):
            if not component.configure():
                self.get_logger().error(f"Failed to configure {tp}: {name}")
                return False
        return True

    def _register_fsm_callbacks(self):
        """Register FSM callbacks, which will be called automatically by the FSM."""
        self.fsm_callbacks = defaultdict(dict)
        # register send action callbacks
        for cb_type, configs in self._config.send_actions.items():
            for key, action in configs.items():
                self.fsm_callbacks[cb_type][key] = partial(
                    self._demonstrator.send_action, action
                )
        # register react callbacks
        # TODO: Allow both prepare before and after.
        for action in DemonstrateAction:
            self.fsm_callbacks[CallbackEventType.PREPARE_EVENT_BEFORE][action] = (
                partial(self._demonstrator.react, action)
            )

    def activate(self) -> bool:
        self._bar = ProgressBar(
            self._sample_limit.size, f"Round {self._sample_info.round}", leave_mode=-1
        )
        Path(self._config.dataset.absolute_directory).mkdir(parents=True, exist_ok=True)
        self.get_logger().info("Warming up...")
        self.capture(warm_up=True)
        return True

    def deactivate(self) -> bool:
        return True

    def sample(self) -> bool:
        """
        Start to sample the data (switch the leaders mode to passive)
        """
        if self.is_reached_round:
            self.get_logger().warning("Maximum number of rounds was reached.")
            return False
        self.get_logger().info(
            Bcolors.green(f"Start sampling round: {self._sample_info.round}")
        )
        self._save_path = self._sampler.compose_path(
            self._config.dataset.absolute_directory, self._sample_info.round
        )
        self._bar.reset(desc=f"Round {self._sample_info.round}")
        return True

    def capture(self, warm_up: bool = False) -> Dict[str, Any]:
        # TODO: can be called when sampling?
        start = time.perf_counter()
        # TODO: configure each action
        data = self._demonstrator.capture_observation(2)
        data = self._config.key_remap(self._config.key_merge(data))
        self._metrics["durations"]["demonstrate/update/demonstrator"] = (
            time.perf_counter() - start
        )
        self.last_capture = data
        # update the visualizers
        start = time.perf_counter()
        for name, visualizer in self._visualizers.items():
            self.get_logger().debug("Updating visualizer %s", name)
            visualizer.update(data, self._sample_info, warm_up)
        self._metrics["durations"]["demonstrate/update/visualizers"] = (
            time.perf_counter() - start
        )
        self._metrics["durations"].update(
            self._demonstrator.metrics.get("durations", {})
        )
        return data

    def update(self) -> bool:
        """
        Update the components (including visualizers).
        """
        # TODO: should react and post action in capture and update?
        info = self._sample_info
        if info.index == 0:
            self.start_stamp = time.perf_counter()
        if self.is_reached:
            self.get_logger().warning(
                f"Sample limitation reached: {info.index} samples"
            )
            return False
        else:
            start = time.perf_counter()
            # TODO: Should use a .copy() to avoid the data being updated in-place within the demonstrator, which could lead to data overwriting issues during asynchronous updating?
            data = self.capture()
            data.update({"log_stamps": time.time_ns()})

            # update the sampler
            def update_sampler(data: dict):
                start_sampler = time.perf_counter()
                for key, value in self._sampler.update(data).items():
                    self._round_data[key].append(value)
                self._metrics["durations"]["demonstrate/update/sampler"] = (
                    time.perf_counter() - start_sampler
                )

            self._submit_action(DemonstrateAction.update, update_sampler, data)
            # update the progress bar
            start_bar = time.perf_counter()
            info.index += 1
            self._bar.update(info.index)
            self._metrics["durations"]["demonstrate/update/bar"] = (
                time.perf_counter() - start_bar
            )
            self._metrics["durations"]["demonstrate/update"] = (
                time.perf_counter() - start
            )
            return True

    def _show_save_info(self, path: str, flag: bool) -> bool:
        if flag:
            self.get_logger().info(Bcolors.green(f"Saved to {path}"))
        else:
            self.get_logger().error(f"Failed to save to {path}")
        return flag

    def save(self) -> None:
        """Save the sampled data and be ready for the next round."""
        self._wait_action_futures(DemonstrateAction.update)
        save_path = self._save_path
        if self._use_executor(DemonstrateAction.save):
            future = self._submit_action(
                DemonstrateAction.save, self._sampler.save, save_path, self._round_data
            )
            future.add_done_callback(
                lambda f: self._show_save_info(save_path, f.result())
            )
        else:
            if not self._show_save_info(
                save_path, self._sampler.save(save_path, self._round_data)
            ):
                return False
        self._sample_info.round += 1
        self._clear()
        return True

    def remove(self) -> bool:
        """Remove the last round saved sample."""
        last_round = self._sample_info.round - 1
        if last_round >= 0:
            path = self._sampler.compose_path(
                self._config.dataset.absolute_directory, last_round
            )
            self._wait_action_futures(DemonstrateAction.save)
            # try to remove the data
            if not self._remove(path, True):
                return False
            # the order is important
            self._sample_info.round -= 1
            self._clear()
            self.get_logger().info(Bcolors.green(f"Removed {path}"))
        else:
            self.get_logger().warning("Not ever saved yet")
        return True

    def _remove(self, path: str, log: bool = False) -> bool:
        removed = self._sampler.remove(path)
        if removed is None:
            self._remove_path(path, log)
            return True
        elif removed:
            return True
        return False

    def _remove_path(self, path: str, log: bool = False) -> bool:
        """Remove the data from the given or last saved path."""
        path_cls = Path(path)
        if path_cls.exists():
            if self._config.remove_mode == "permanent":
                if path_cls.is_dir():
                    shutil.rmtree(path_cls)
                else:
                    path_cls.unlink()
            else:
                send2trash(path_cls)
            return True
        else:
            if log:
                self.get_logger().warning(f"Path to be removed {path} does not exist.")
            return True

    def _clear(self) -> None:
        self._round_data = defaultdict(list)
        self._sampler.clear()
        self._sample_info.index = 0
        self._save_path = ""

    def _use_executor(self, action: DemonstrateAction) -> bool:
        return action in self._action_executors

    def _submit_action(
        self, action: DemonstrateAction, func: Any, *args, **kwargs
    ) -> Future:
        future = self._action_executors[action].submit(func, *args, **kwargs)
        self._action_futures[action].append(future)
        return future

    def _cancel_action_futures(self, action: DemonstrateAction) -> None:
        futures = self._action_futures.get(action, None)
        if futures:
            for future in futures:
                future.cancel()
            self._action_futures[action] = []

    def _wait_action_futures(self, action: DemonstrateAction) -> None:
        futures = self._action_futures.get(action, None)
        if futures:
            for update_future in tqdm(
                as_completed(futures), f"Completing {action.name} futures", len(futures)
            ):
                update_future.result()
            self._action_futures[action] = []

    def abandon(self) -> bool:
        """Abandon the current round of sampling."""
        self._cancel_action_futures(DemonstrateAction.update)
        self._remove_path(self._save_path, False)
        self._clear()
        self.get_logger().info(
            Bcolors.green(f"Abandoned the current round: {self._sample_info.round}")
        )
        return True

    def finish(self) -> bool:
        """
        Finish the demonstration.
        """
        self.get_logger().info(
            f"Finished the demonstration: from {self._sample_limit.start_round} to {self._sample_info.round}"
        )
        for vis in self._visualizers.values():
            vis.shutdown()
        self._bar.close()
        return True

    def log_round(self):
        self.get_logger().info(
            Bcolors.cyan(f"Current sample round: {self._sample_info.round}")
        )

    @property
    def is_reached(self) -> bool:
        limit = self._sample_limit
        reach_size = limit.size > 0 and self._sample_info.index >= limit.size
        reach_duration = (
            limit.duration > 0
            and time.perf_counter() - self.start_stamp >= limit.duration
        )
        return reach_size or reach_duration

    @property
    def is_reached_round(self) -> bool:
        end_round = self._sample_limit.end_round
        return end_round > 0 and self._sample_info.round > end_round

    @property
    def demonstrator(self) -> Demonstrator:
        return self._demonstrator

    @property
    def metrics(self) -> Dict[str, Any]:
        return self._metrics



import asyncio
import numpy as np
from threading import Event
from time import time_ns
from typing import Union, Optional
from linuxpy.video.device import Capability, Device, PixelFormat, VideoCapture
from turbojpeg import TurboJPEG
from pydantic import field_validator, model_validator, ValidationInfo
from airbot_data_collection.common.systems.basis import Sensor, DictDataStamped
from airbot_data_collection.common.devices.cameras.utils import (
    ColorCameraConfig,
    find_camera_indices,
    get_camera_index_by_bus_info,
    CameraInfo,
    CameraControl,
)
from airbot_data_collection.common.visualizers.basis import VisualizerBasis
from airbot_data_collection.common.utils.progress import run_event_loop
from airbot_data_collection.common.utils.codec import ImageCoder


class V4L2CameraConfig(ColorCameraConfig):
    """Configuration for V4L2 camera device."""

    nb_buffers: int = 2
    mode: Optional[Union[str, int]] = None
    decode: bool = True
    pixel_format: PixelFormat = PixelFormat.MJPEG

    @field_validator("mode", mode="after")
    def validate_mode(cls, v):
        return {
            "mmap": Capability.STREAMING,
            "read": Capability.READWRITE,
        }.get(v, v)

    @field_validator("pixel_format", mode="before")
    def validate_pixel_format(cls, v):
        if isinstance(v, str):
            return PixelFormat[v.upper()]
        return v

    @model_validator(mode="after")
    def validate_rgb_camera(self):
        if self.rgb_camera.pixel_format is None:
            object.__setattr__(self.rgb_camera, "pixel_format", self.pixel_format)
        return self

    @property
    def color_format(self):
        color_cfg = self.rgb_camera
        return (color_cfg.width, color_cfg.height, color_cfg.pixel_format)

    @property
    def can_set_format(self) -> bool:
        return None not in self.color_format

    @property
    def is_partial_color_format(self) -> bool:
        formats = set(self.color_format)
        return None in formats and len(formats) > 1


class V4L2Camera(Sensor):
    """
    V4L2 camera class for Linux systems.
    """

    def __init__(self, config: V4L2CameraConfig):
        self.config = config
        self._shutdown = False
        self._visualizer = None
        self._frame = None

    def on_configure(self) -> bool:
        config = self.config
        color_config = config.rgb_camera
        cam_id = config.camera_index
        if cam_id is None:
            cam_id = find_camera_indices()[0]
        if isinstance(cam_id, int) or cam_id.isdigit():
            self.device = Device.from_id(int(cam_id))
        else:
            if "usb" in cam_id:
                cam_id = get_camera_index_by_bus_info(cam_id)[0]
            self.device = Device(cam_id)
        self.device.open()
        if self.device.closed:
            return False
        self._capture = VideoCapture(self.device, config.nb_buffers, config.mode)
        if config.can_set_format:
            self._capture.set_format(
                color_config.width, color_config.height, color_config.pixel_format
            )
        elif config.is_partial_color_format:
            self.get_logger().warning(
                "Partial color format specified, but cannot set format without all parameters."
            )
        self._format = self._capture.get_format()
        if color_config.fps:
            self._capture.set_fps(color_config.fps)
        # NOTE: do not move to __init__ to avoid deepcopy error
        self._event = Event()
        self._read_fut = asyncio.run_coroutine_threadsafe(
            self._read_frame(), run_event_loop()
        )
        if self.config.decode and color_config.pixel_format is PixelFormat.MJPEG:
            self._jpeg = TurboJPEG()
        self._init_info()
        return True

    def capture_observation(
        self, timeout: Optional[float] = None
    ) -> DictDataStamped[Union[bytes, np.ndarray]]:
        if self.config.blocking or self._frame is None:
            if not self._event.wait(timeout):
                raise TimeoutError(f"Timeout waiting for camera frame: {timeout} s.")
            self._event.clear()
        frame_bytes = bytes(self._frame)
        key = "color/image_raw"
        obs = {key: {"t": self.stamp}}
        if not self.config.decode:
            obs[key]["data"] = frame_bytes
        else:
            pixel_format = self._format.pixel_format
            if pixel_format is PixelFormat.MJPEG:
                image = self._jpeg.decode(frame_bytes)
                if image.shape[0] == 0 or image.shape[1] == 0:
                    raise ValueError("Received empty image from camera.")
            elif pixel_format is PixelFormat.YUYV:
                image = ImageCoder.yuyv2bgr(
                    frame_bytes, self._format.width, self._format.height
                )
            else:
                raise NotImplementedError(
                    f"Pixel format {pixel_format} not supported for decoding yet."
                )
            color_config = self.config.rgb_camera
            if color_config.color_mode == "rgb":
                image = image[:, :, ::-1]
            obs[key]["data"] = image
        return obs

    def shutdown(self) -> bool:
        # TODO: why manually closing raises error?
        # self._shutdown = True
        # self._capture.close()
        # self._read_fut.result()
        # # self.device.close()
        # return self.device.closed
        return True

    def _init_info(self):
        cam_format = self._capture.get_format()
        camera_info = CameraInfo(
            width=cam_format.width, height=cam_format.height
        ).model_dump(mode="json")
        config = self.config.rgb_camera
        if config.intrinsics is not None:
            camera_info.update(config.intrinsics.model_dump(mode="json"))
        if config.calibration is not None:
            camera_info.update(config.calibration.model_dump(mode="json"))
        color_info = {"camera_info": camera_info}
        self.device.controls._init_if_needed()
        id_to_name = {}
        for ctrl in self.device.info.controls:
            id_to_name[ctrl.id] = ctrl.name.decode().lower().replace(" ", "_")
        ctrl_info = {}
        for key, ctrl in self.device.controls.items():
            ctrl_info[id_to_name[key]] = ctrl.value
        fps = self._capture.get_fps().as_integer_ratio()
        color_info.update(
            {"fps": str(fps[0] / fps[1]), "pixel_format": cam_format.pixel_format.name}
        )
        color_info.update(ctrl_info)
        dev_info = self.device.info
        self._info = {
            "driver": dev_info.driver,
            "card": dev_info.card,
            "bus_info": dev_info.bus_info,
            "version": dev_info.version,
            "color": color_info,
        }

    def get_info(self):
        return self._info

    def set_visualizer(self, visualizer: VisualizerBasis, prefix: str = ""):
        """TODO: should use this method?
        Set the visualizer for this camera for self control.
        :param visualizer: VisualizerBasis instance to visualize the camera data.
        :param prefix: Prefix for the visualizer key.
        """
        self._visualizer = visualizer

    async def _read_frame(self):
        with self._capture as stream:
            async for frame in stream:
                self.stamp = time_ns()
                self._frame = frame
                self._event.set()
                # if self._shutdown:
                #     break


if __name__ == "__main__":
    import time
    import cv2

    camera = V4L2Camera()
    assert camera.configure()
    while True:
        start = time.monotonic()
        image = camera.capture_observation()
        print(f"time cost: {time.monotonic() - start}s", end="\r")
        cv2.imshow("image", image)
        if cv2.waitKey(1) & 0xFF in (ord("q"), 27):  # 27 is the ESC key
            break
    assert camera.shutdown()



from airbot_data_collection.state_machine.basis import ToDestConfig
from airbot_data_collection.state_machine.fsm import (
    DemonstrateAction as Action,
    DemonstrateState as State,
    StateMachineConfig,
)


STATE_MACHINE_CONFIG = StateMachineConfig(
    states=State,
    initial=State.unconfigured,
    action_transitions={
        Action.configure: {
            State.unconfigured: [
                ToDestConfig(dest=State.inactive),
            ],
        },
        Action.activate: {
            State.inactive: [
                ToDestConfig(dest=State.active),
            ]
        },
        Action.sample: {
            # a prepare callback with the same name as the action
            # will be automatically added to the
            # first transition of each source and the result of
            # the callback will be used as the conditions
            # for the first None conditions & unless transition and
            # and as the unless for the last None conditions & unless
            # transition
            State.active: [
                # success
                # the after callback will use the corresponding trigger
                ToDestConfig(dest=State.sampling),
                # # failure
                # # the before callback will use the original method
                # ToDestConfig(dest=None),
            ],
        },
        Action.update: {
            State.sampling: [
                # success
                ToDestConfig(dest=None),
                # failure
                ToDestConfig(dest=None),
            ],
        },
        Action.abandon: {
            State.sampling: [
                ToDestConfig(dest=State.active),
            ]
        },
        Action.save: {
            State.sampling: [
                ToDestConfig(dest=State.active),
            ]
        },
        Action.remove: {State.active: [ToDestConfig(dest=None)]},
        Action.finish: {State.active: [ToDestConfig(dest=State.finalized)]},
        Action.capture: {
            State.active: [ToDestConfig(dest=None)],
            State.sampling: [ToDestConfig(dest=None)],
        },
    },
)


def get_fsm_config(**kwargs) -> StateMachineConfig:
    if kwargs:
        return StateMachineConfig.model_validate(
            STATE_MACHINE_CONFIG.model_copy(update=kwargs)
        )
    return STATE_MACHINE_CONFIG



def main():
    """示例: 同时查看多台 RealSense 相机的彩色与对齐深度图。

    用法:
        python test_realsense_depth.py                        # 自动发现全部设备
        python test_realsense_depth.py -d 12345678 87654321    # 指定多台序列号(空格分隔)
        python test_realsense_depth.py -d 12345678 -a          # 只指定一台并开启对齐
        python test_realsense_depth.py --headless             # 无界面运行(仅打印 FPS)

    增强点:
        1. --device_id 现在支持 0/多参数 (nargs="*")；不提供则自动发现全部设备。
        2. 多设备并行: 为每台设备建立独立 pipeline + align (可选)。
        3. 将所有设备的 (color | depth_colormap) 按列拼接，再把多设备行上下堆叠显示。
    4. 支持 --headless 无窗口模式; GUI 模式按 q 退出。
    """
    import argparse
    import cv2
    import numpy as np
    import pyrealsense2 as rs
    import time
    from collections import defaultdict

    parser = argparse.ArgumentParser(description="Intel RealSense 多设备深度测试")
    parser.add_argument(
        "-d",
        "--device_id",
        nargs="*",
        help="一个或多个相机序列号(空格分隔)。缺省: 自动发现全部已连接设备。",
    )
    parser.add_argument(
        "-a",
        "--align",
        action="store_true",
        help="将深度帧对齐到彩色帧 (默认关闭)",
    )
    parser.add_argument(
        "-wd", "--width", type=int, default=640, help="彩色/深度流宽度 (默认640)"
    )
    parser.add_argument(
        "-hi", "--height", type=int, default=480, help="彩色/深度流高度 (默认480)"
    )
    parser.add_argument("--fps", type=int, default=30, help="帧率 (默认30)")
    parser.add_argument(
        "-hl",
        "--headless",
        action="store_true",
        help="无界面模式: 不显示窗口, 周期性打印帧率",
    )
    parser.add_argument(
        "--log-interval",
        type=float,
        default=2.0,
        help="headless 模式下打印 FPS 的时间间隔(秒, 默认2.0)",
    )
    parser.add_argument(
        "-st",
        "--stack",
        action="store_true",
        help="将多台设备的图像按列堆叠显示 (默认关闭)",
    )
    parser.add_argument(
        "-dc",
        "--depth-color",
        action="store_true",
        help="将深度图像应用伪彩色 (默认关闭)",
    )
    args = parser.parse_args()

    # 发现设备
    ctx = rs.context()
    connected = [d.get_info(rs.camera_info.serial_number) for d in ctx.query_devices()]
    if not connected:
        print("未发现任何 RealSense 设备。")
        return

    if args.device_id and len(args.device_id) > 0:
        # 过滤: 只保留实际存在的
        device_ids = []
        for did in args.device_id:
            if did in connected:
                device_ids.append(did)
            else:
                print(f"警告: 指定的设备 {did} 未连接, 忽略。")
        if not device_ids:
            print("未找到任一指定设备, 退出。")
            return
    else:
        device_ids = connected

    print(f"将使用设备: {device_ids}")

    width, height, fps = args.width, args.height, args.fps
    align_enabled = args.align
    headless = args.headless

    # 为每台设备构建 pipeline
    pipelines = []  # list[(serial, pipeline, align_or_None)]
    try:
        for serial in device_ids:
            pipeline = rs.pipeline()
            cfg = rs.config()
            cfg.enable_device(serial)
            cfg.enable_stream(rs.stream.color, width, height, rs.format.bgr8, fps)
            cfg.enable_stream(rs.stream.depth, width, height, rs.format.z16, fps)
            print(f"启动设备 {serial}...")
            profile = pipeline.start(cfg)
            aligner = rs.align(rs.stream.color) if align_enabled else None
            # 预热: 丢弃前几帧
            for _ in range(5):
                pipeline.wait_for_frames()
            pipelines.append((serial, pipeline, aligner))
            time.sleep(0.05)

        if headless:
            print("全部设备已启动 (headless 模式)。按 Ctrl+C 结束。")
        else:
            print("全部设备已启动, 按 q / ESC 退出。")

        # headless 统计
        per_device_counts = {serial: 0 for serial, *_ in pipelines}
        images = defaultdict(dict)
        last_log = time.perf_counter()

        while True:
            per_device_cols = [] if not headless else None
            now = time.perf_counter()
            for serial, pipeline, aligner in pipelines:
                try:
                    frames = pipeline.wait_for_frames(timeout_ms=1000)
                except Exception as e:  # noqa: BLE001
                    print(f"设备 {serial} 获取帧失败: {e}")
                    continue
                if aligner is not None:
                    frames = aligner.process(frames)
                color_frame = frames.get_color_frame()
                depth_frame = frames.get_depth_frame()
                if not color_frame or not depth_frame:
                    continue
                per_device_counts[serial] += 1

                if headless:
                    # 仅统计，不做图像处理
                    continue

                color_img = np.asanyarray(color_frame.get_data())
                depth_img = np.asanyarray(depth_frame.get_data())

                if not headless:
                    if args.depth_color:
                        depth_color = cv2.applyColorMap(
                            cv2.convertScaleAbs(depth_img, alpha=0.03), cv2.COLORMAP_JET
                        )
                        if color_img.shape[:2] != depth_color.shape[:2]:
                            depth_color = cv2.resize(
                                depth_color, (color_img.shape[1], color_img.shape[0])
                            )
                    else:
                        depth_color = depth_img
                        if args.stack:
                            raise NotImplementedError(
                                "堆叠彩色图和纯深度图的显示功能尚未实现"
                            )
                    # cv2.putText(
                    #     color_img,
                    #     f"{serial}",
                    #     (10, 25),
                    #     cv2.FONT_HERSHEY_SIMPLEX,
                    #     0.8,
                    #     (0, 255, 0),
                    #     2,
                    # )
                    if args.stack:
                        col = np.vstack((color_img, depth_color))
                        per_device_cols.append(col)
                    else:
                        images[serial]["color"] = color_img
                        images[serial]["depth"] = depth_img

            # headless 日志输出
            if (now - last_log) >= args.log_interval:
                elapsed = now - last_log
                fps_info = [
                    f"{serial}:{per_device_counts[serial] / elapsed:.1f}fps"
                    for serial in per_device_counts
                ]
                print(" | ".join(fps_info))
                for k in per_device_counts:
                    per_device_counts[k] = 0
                last_log = now

            if headless:
                continue

            if args.stack:
                if not per_device_cols:
                    continue
                canvas = np.hstack(per_device_cols)
                cv2.imshow("RealSense Multi-Depth", canvas)
            else:
                for serial, imgs in images.items():
                    color_img = imgs.get("color")
                    depth_img = imgs.get("depth")
                    if color_img is not None and depth_img is not None:
                        cv2.imshow(f"RealSense {serial} Color", color_img)
                        cv2.imshow(f"RealSense {serial} Depth", depth_img)
            if cv2.waitKey(1) & 0xFF in {ord("q"), 27}:
                break
    except (KeyboardInterrupt, NotImplementedError):
        print("用户中断, 退出。")
    finally:
        for serial, pipeline, _ in pipelines:
            print(f"停止设备 {serial}")
            try:
                pipeline.stop()
            except Exception:  # noqa: BLE001
                pass
        cv2.destroyAllWindows()


if __name__ == "__main__":
    main()



import numpy as np
from tkinter import Canvas, Tk, Toplevel
from typing import Union
from PIL import Image, ImageTk
from airbot_data_collection.common.visualizers.basis import (
    GUIVisualizerConfig,
    SampleInfo,
    VisualizerBasis,
)
from airbot_data_collection.basis import DictDataStamped


def get_dpi() -> float:
    root = Tk()
    dpi = root.winfo_fpixels("1i")  # 水平方向的DPI
    root.destroy()
    return dpi


def resolution_to_inches(width: int, height: int) -> tuple[float, float]:
    dpi = get_dpi()
    return width / dpi, height / dpi


class TkinterVisualizer(VisualizerBasis):
    config: GUIVisualizerConfig

    def on_configure(self):
        # 创建主窗口，但仅用于显示信息文本
        self.root = Tk()
        self.root.title(self.config.title or self.__class__.__name__)
        # 设置主窗口的Canvas高度为40，仅显示信息
        self.canvas = Canvas(self.root, width=self.config.width, height=40, bg="gray")
        self.canvas.pack()
        self.text_ids = {}
        self.windows: dict[str, dict[str, Union[Canvas, Toplevel]]] = {}
        return True

    def on_update(self, data: DictDataStamped[np.ndarray], info: SampleInfo):
        for title, data_dict in data.items():
            img = data_dict["data"]
            image_pil = Image.fromarray(img.astype(np.uint8), mode="RGB")
            img_width, img_height = image_pil.size
            if title not in self.windows:
                window = Toplevel(self.root)
                window.title(title)
                canvas = Canvas(window, width=img_width, height=img_height)
                canvas.pack()
                tk_image = ImageTk.PhotoImage(image_pil)
                image_id = canvas.create_image(0, 0, anchor="nw", image=tk_image)
                self.windows[title] = {
                    "window": window,
                    "canvas": canvas,
                    "image_id": image_id,
                    "tk_image": tk_image,
                    "width": img_width,
                    "height": img_height,
                }
            else:
                win_info = self.windows[title]
                if img_width != win_info["width"] or img_height != win_info["height"]:
                    win_info["canvas"].config(width=img_width, height=img_height)
                    win_info["window"].geometry(f"{img_width}x{img_height}")
                    win_info["width"] = img_width
                    win_info["height"] = img_height
                new_tk_image = ImageTk.PhotoImage(image_pil)
                win_info["canvas"].itemconfig(win_info["image_id"], image=new_tk_image)
                win_info["tk_image"] = new_tk_image

        if not self.config.ignore_info:
            info_text = f"Round: {info.round} | Index: {info.index}"
            if "info" in self.text_ids:
                self.canvas.itemconfig(self.text_ids["info"], text=info_text)
            else:
                self.text_ids["info"] = self.canvas.create_text(
                    10,
                    10,
                    text=info_text,
                    fill="black",
                    font=("Helvetica", 12),
                    anchor="nw",
                )
        self.root.update_idletasks()
        self.root.update()

    def shutdown(self) -> bool:
        for win_info in self.windows.values():
            win_info["window"].destroy()
        self.root.destroy()
        return True


if __name__ == "__main__":
    import time

    def create_color_image(
        color: tuple[int, int, int], shape=(480, 640, 3)
    ) -> np.ndarray:
        return np.ones(shape, dtype=np.uint8) * np.array(color, dtype=np.uint8)

    vis = TkinterVisualizer(GUIVisualizerConfig())
    assert vis.configure()

    images = {
        "Red": create_color_image((255, 0, 0)),
        "Green": create_color_image((0, 255, 0)),
        "Blue": create_color_image((0, 0, 255)),
    }

    round_num = 1
    index = 0

    while True:
        start = time.monotonic()
        vis.update(images, SampleInfo(round=round_num, index=index))
        print(f"time cost: {time.monotonic() - start:.3f}s")
        time.sleep(1)
        index += 1



import os
import sys
import logging
from importlib import import_module
from typing import TYPE_CHECKING, Optional, Dict, Any
from pathlib import Path
from contextlib import suppress


logger = logging.getLogger("ros_utils")


def find_ros_distro(path: str) -> Optional[str]:
    path_obj = Path(path)

    parts = path_obj.parts
    with suppress(ValueError, IndexError):
        ros_index = parts.index("ros")
        return parts[ros_index + 1]


# The Python path may contain multiple versions of ROS,
# causing subsequent program processing errors.
ROS_DISTRO = os.environ["ROS_DISTRO"]
ros_paths = []
for path in sys.path.copy():
    if path.endswith("site-packages") or path.endswith("dist-packages"):
        ros_distro = find_ros_distro(path)
        if ros_distro is not None:
            if ros_distro != ROS_DISTRO:
                logger.warning(f"Removing {path} from sys.path for ROS distro mismatch")
                sys.path.remove(path)
            else:
                ros_paths.append(path)
logger.info(f"Using ROS distro: {ROS_DISTRO} with path: {ros_paths}")
if len(ros_paths) == 0:
    raise ImportError("No ROS paths found")
elif len(ros_paths) > 1:
    logger.warning("Multiple ROS paths found")

ROS_VERSION = os.environ["ROS_VERSION"]
if TYPE_CHECKING:

    def build_short_to_full_msg_map(preferred_packages) -> dict: ...
    def get_message(identifier: str) -> type: ...
    def set_message_fields(
        msg, values, expand_header_auto=False, expand_time_now=False
    ) -> list: ...
    def get_fields_and_field_types(msg) -> Dict[str, str]: ...
    def time_ns_to_stamp(time_ns: int) -> Any: ...
    def stamp_to_time_ns(stamp: Any) -> int: ...
    def get_datatype_and_msgdef_text(msg) -> tuple: ...
    def process_camera_info_dict(cam_info_dict: Dict[str, Any]): ...
    def get_current_stamp() -> Any: ...
else:
    module = import_module(f"airbot_data_collection.common.utils.ros.ros{ROS_VERSION}")
    build_short_to_full_msg_map = module.build_short_to_full_msg_map
    get_message = module.get_message
    set_message_fields = module.set_message_fields
    get_fields_and_field_types = module.get_fields_and_field_types
    time_ns_to_stamp = module.time_ns_to_stamp
    stamp_to_time_ns = module.stamp_to_time_ns
    get_datatype_and_msgdef_text = module.get_datatype_and_msgdef_text
    process_camera_info_dict = getattr(
        module, "process_camera_info_dict", lambda x: None
    )
    get_current_stamp = module.get_current_stamp


MSG_MAP: dict = {}


def get_message_short(identifier: str, cache: bool = True) -> Optional[type]:
    global MSG_MAP
    if cache:
        if not MSG_MAP:
            preferred = ("std_msgs", "geometry_msgs", "sensor_msgs")
            if ROS_VERSION != "1":
                preferred += ("builtin_interfaces",)
            MSG_MAP = build_short_to_full_msg_map(preferred)
        full_identifier = MSG_MAP.get(identifier)
    else:
        full_identifier = build_short_to_full_msg_map(preferred).get(identifier)
    if full_identifier is not None:
        return get_message(full_identifier)


if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.INFO)

    # if ROS_VERSION == "1":
    #     import rospy

    #     rospy.init_node("test_set_message_fields")

    # mapping = build_short_to_full_msg_map()
    # print(f"Total messages found: {len(mapping)}")
    # from pprint import pprint

    # pprint(mapping)

    # from std_msgs.msg import String

    # msg = String()
    # set_message_fields(msg, {"data": "Hello"})
    # assert msg.data == "Hello", msg.data

    # from geometry_msgs.msg import PointStamped

    # msg = PointStamped()
    # setters = set_message_fields(msg, {"header": "auto"}, expand_header_auto=True)
    # for s in setters:
    #     s()  # sets msg.header.stamp = rospy.Time.now()
    # print(msg.header.stamp)
    # from geometry_msgs.msg import PointStamped

    # msg = PointStamped()
    # setters = set_message_fields(
    #     msg,
    #     {"header": {"stamp": "now"}, "point": {"x": 1.0, "y": 2.0, "z": 0.0}},
    #     expand_time_now=True,
    # )

    # # Apply time now
    # for s in setters:
    #     s()

    # print(msg.header.stamp)  # Current time

    # assert get_message_short("PointStamped") == PointStamped

    # pprint(get_fields_and_field_types(PointStamped()))

    # ns = 156789123456789
    # stamp = time_ns_to_stamp(ns)
    # print(stamp)
    # assert stamp_to_time_ns(stamp) == ns

    # from std_msgs.msg import String

    # datatype, msgdef = get_datatype_and_msgdef_text(String)
    # print(datatype, msgdef)

    pass



import numpy as np
from typing import Any, Generic, TypeVar
from abc import abstractmethod
from pydantic import BaseModel, ConfigDict
from logging import getLogger
from airbot_data_collection.basis import ConfigurableBasis


T = TypeVar("T")


class EnvironmentOutput(BaseModel, Generic[T]):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # The current observation from the environment
    observation: T
    # The reward obtained from the last action
    reward: float = 0.0
    # Whether the agent reaches the terminal state (as defined under the MDP of the task) which can be positive or negative. If true, the user needs to call reset().
    terminated: bool = False
    # Whether the truncation condition outside the scope of the MDP is satisfied. Typically, this is a timelimit, but could also be used to indicate an agent physically going out of bounds. Can be used to end the episode prematurely before a terminal state is reached. If true, the user needs to call reset().
    truncated: bool = False
    # Contains auxiliary diagnostic information (helpful for debugging, learning, and logging). This might, for instance, contain: metrics that describe the agent’s performance state, variables that are hidden from observations, or individual reward terms that are combined to produce the total reward.
    info: dict = {}


class EnvironmentBasis(ConfigurableBasis, Generic[T]):
    @abstractmethod
    def reset(self) -> None:
        """Reset the environment"""

    @abstractmethod
    def input(self, input: Any) -> None:
        """Take an action in the environment"""

    @abstractmethod
    def output(self) -> EnvironmentOutput[T]:
        """Get the current output from the environment.
        This method should not change the state of the environment.
        """

    @abstractmethod
    def shutdown(self) -> bool:
        """Shutdown the environment"""

    @classmethod
    def get_logger(cls):
        return getLogger(cls.__name__)


class MockEnvironmentConfig(BaseModel):
    max_steps: int = 0
    output: Any = None


class MockEnvironment(EnvironmentBasis):
    config: MockEnvironmentConfig

    def on_configure(self) -> bool:
        self.reset()
        return True

    def reset(self) -> None:
        self.get_logger().info("Environment resetting...")
        self._t = 0
        self._output_t = 0

    def input(self, input: Any) -> None:
        # self.get_logger().info(f"Environment input: {input}")
        self._t += 1

    def output(self) -> EnvironmentOutput[np.ndarray]:
        # self._output_t += 1
        # if self._t != self._output_t:
        #     raise RuntimeError(
        #         "The environment output is out of sync with the input. Please make sure to call input() before output()"
        #     )
        terminated = (
            (self._t >= self.config.max_steps) if self.config.max_steps > 0 else False
        )
        # self.get_logger().info(f"Environment step: {self._t}, terminated: {terminated}")
        return EnvironmentOutput(
            observation=np.array([self._t])
            if self.config.output is None
            else self.config.output,
            terminated=terminated,
        )

    def shutdown(self) -> bool:
        return True



from datetime import datetime, timezone
from collections import defaultdict
from functools import wraps
import os
import sys
import locale


def inside_slurm():
    """Check whether the python process was launched through slurm"""
    # TODO(rcadene): return False for interactive mode `--pty bash`
    return "SLURM_JOB_ID" in os.environ


def format_big_number(num, precision=0):
    suffixes = ["", "K", "M", "B", "T", "Q"]
    divisor = 1000.0

    for suffix in suffixes:
        if abs(num) < divisor:
            return f"{num:.{precision}f}{suffix}"
        num /= divisor

    return num


def capture_timestamp_utc():
    return datetime.now(timezone.utc)


def check_utf8_locale() -> bool:
    lang = os.environ.get("LANG", None)
    if lang is None or "UTF-8" in lang:
        encoding = sys.getdefaultencoding()
        if encoding == "utf-8":
            return True
        else:
            sys.stderr.write(f"Python default encoding is not UTF-8: {encoding}\n")
    else:
        sys.stderr.write(f"System default locale is not UTF-8: {lang}\n")
    return False


def set_utf8_locale() -> bool:
    if check_utf8_locale():
        sys.stderr.write("UTF-8 locale is already set\n")
        return True

    os.environ["PYTHONIOENCODING"] = "utf-8"
    os.environ["LANG"] = "en_US.UTF-8"
    os.environ["LC_ALL"] = "en_US.UTF-8"

    for lc in ["zh_CN.UTF-8", "en_US.UTF-8", "C.UTF-8"]:
        try:
            locale.setlocale(locale.LC_ALL, lc)
            sys.stderr.write(f"Locale set to '{lc}'\n")
            return True
        except locale.Error as e:
            sys.stderr.write(f"Warning: Could not set locale to '{lc}': {e}\n")
    sys.stderr.write("Failed to set UTF-8 locale\n")
    return False


def defaultdict_to_dict(d: defaultdict):
    if isinstance(d, defaultdict):
        d = {k: defaultdict_to_dict(v) for k, v in d.items()}
    elif isinstance(d, dict):
        d = {k: defaultdict_to_dict(v) for k, v in d.items()}
    elif isinstance(d, list):
        d = [defaultdict_to_dict(x) for x in d]
    return d


def ensure_equal_length(ref: list, value: list, one_copy: bool = True):
    if len(value) == 1 and one_copy:
        value *= len(ref)
    if len(ref) != len(value):
        raise ValueError(f"Length mismatch: {len(ref)} vs {len(value)}")
    return value


def unpacking(func):
    """Decorator to unpack arguments from a single list or tuple"""

    @wraps(func)
    def wrapper(args):
        return func(*args)

    return wrapper
